#!/usr/bin/env python

"""Extends requests.Session to enable simple checksum testing."""

__all__ = ["ValidatingRequestsSession"]

import hashlib

import requests

from .exceptions import ChecksumFailed


class ValidatingRequestsSession(requests.Session):
    """Download a file and test whether it matches a checksum."""

    def __init__(self, *args, checksum_algorithm=hashlib.sha256, **kwargs):
        """
        Download a file and test whether it matches a checksum.

        Arguments
        ---------
        checksum_algorithm : function
            algorithm to use to create checksum of downloaded file,
            default: hashlib.sha256
        *args, **kwargs
            any argument accepted by `requests.Session`
        """
        super().__init__(*args, **kwargs)
        self._algorithm = checksum_algorithm

    def get(self, url, checksum, **kwargs):
        """Send a GET request, tests checksum."""
        kwargs.setdefault("allow_redirects", True)
        return self.request("GET", url, checksum, **kwargs)

    def post(self, url, checksum, **kwargs):
        """Send a POST request, tests checksum."""
        return self.request("POST", url, checksum, **kwargs)

    # delete, put, head don’t return data,
    # testing checksums does not apply

    def request(self, method, url, checksum, **kwargs):
        """
        Retrieve file from cache or proxy requests.request.

        Raise `ChecksumFailed` if the file’s digest and `checksum` differ.
        """
        response = super().request(method, url, **kwargs)
        digest = self._algorithm(response.content).hexdigest()

        if digest != checksum:
            raise ChecksumFailed(
                f"Checksum failed for {url}, expected {checksum}, got {digest}"
            )

        return response



#!/usr/bin/env python3


"""Find detailed routes between two points."""


import copy
import collections
import datetime
import functools
import warnings

import jpype
import pyproj
import shapely

from .access_leg import AccessLeg
from .direct_leg import DirectLeg
from .egress_leg import EgressLeg
from .street_segment import StreetSegment
from .transfer_leg import TransferLeg
from .transit_leg import TransitLeg
from .transport_mode import TransportMode
from .trip import Trip
from ..util import GoodEnoughEquidistantCrs, start_jvm

import com.conveyal.r5
import gnu.trove.map
import java.lang


__all__ = ["TripPlanner"]


start_jvm()


ACCURATE_GEOMETRIES = com.conveyal.r5.transit.TransitLayer.SAVE_SHAPES
COORDINATE_CORRECTION_FACTOR = com.conveyal.r5.streets.VertexStore.FIXED_FACTOR
R5_CRS = "EPSG:4326"

ONE_MINUTE = datetime.timedelta(minutes=1)
ZERO_SECONDS = datetime.timedelta(seconds=0)


class TripPlanner:
    """Find detailed routes between two points."""

    MAX_ACCESS_TIME = datetime.timedelta(hours=1)
    MAX_EGRESS_TIME = MAX_ACCESS_TIME

    def __init__(self, transport_network, request):
        """
        Find detailed routes between two points.

        Arguments
        =========
        transport_network : r5py.r5.TransportNetwork
            A transport network to route on
        request : r5py.r5.regional_task
            The parameters that should be used when finding a route
        """
        self.transport_network = transport_network
        self.request = request

        EQUIDISTANT_CRS = GoodEnoughEquidistantCrs(self.transport_network.extent)
        self._crs_transformer_function = pyproj.Transformer.from_crs(
            R5_CRS,
            EQUIDISTANT_CRS,
            always_xy=True,
        ).transform

    @property
    def trips(self):
        """
        Detailed routes between two points.

        Returns
        =======
        list[r5py.r5.Trip]
            Detailed routes that meet the requested parameters
        """
        trips = self.direct_paths + self.transit_paths
        return trips

    @property
    def direct_paths(self):
        """
        Detailed routes between two points using direct modes.

        Returns
        =======
        list[r5py.r5.Trip]
            Detailed routes that meet the requested parameters, using direct
            modes (walking, cycling, driving).
        """
        direct_paths = []
        request = copy.copy(self.request)

        direct_modes = [mode for mode in request.transport_modes if mode.is_street_mode]

        for transport_mode in direct_modes:
            # short-circuit identical from_id and to_id:
            if (
                request._regional_task.fromLat == request._regional_task.toLat
                and request._regional_task.fromLon == request._regional_task.toLon
            ):
                lat = request._regional_task.fromLat
                lon = request._regional_task.fromLon
                direct_paths.append(
                    Trip(
                        [
                            DirectLeg(
                                transport_mode,
                                collections.namedtuple(
                                    "StreetSegment",
                                    ["distance", "duration", "geometry"],
                                )(0.0, 0.0, f"LINESTRING({lon} {lat}, {lon} {lat})"),
                            )
                        ]
                    )
                )
            else:
                street_router = com.conveyal.r5.streets.StreetRouter(
                    self.transport_network.street_layer
                )
                street_router.profileRequest = request
                street_router.streetMode = transport_mode

                street_router.setOrigin(
                    request._regional_task.fromLat,
                    request._regional_task.fromLon,
                )
                street_router.setDestination(
                    request._regional_task.toLat,
                    request._regional_task.toLon,
                )

                street_router.route()

                try:
                    router_state = street_router.getState(
                        street_router.getDestinationSplit()
                    )
                    street_segment = self._street_segment_from_router_state(
                        router_state,
                        transport_mode,
                    )
                    direct_paths.append(
                        Trip(
                            [
                                DirectLeg(transport_mode, street_segment),
                            ]
                        )
                    )
                except (
                    java.lang.NullPointerException,
                    java.util.NoSuchElementException,
                ):
                    warnings.warn(
                        f"Could not find route between origin "
                        f"({self.request._regional_task.fromLon}, "
                        f"{self.request._regional_task.fromLat}) "
                        f"and destination ({self.request._regional_task.toLon}, "
                        f"{self.request._regional_task.toLat})",
                        RuntimeWarning,
                    )
        return direct_paths

    def _street_segment_from_router_state(self, router_state, transport_mode):
        """Retrieve a StreetSegment for a route."""
        street_path = com.conveyal.r5.profile.StreetPath(
            router_state,
            self.transport_network,
            False,
        )
        street_segment = StreetSegment(street_path)
        return street_segment

    @functools.cached_property
    def transit_paths(self):
        """
        Detailed routes between two points on public transport.

        Returns
        =======
        list[r5py.r5.Trip]
            Detailed routes that meet the requested parameters, on public
            transport.
        """
        transit_paths = []

        # if any transit mode requested:
        if [mode for mode in self.request.transport_modes if mode.is_transit_mode]:
            request = copy.copy(self.request)

            midnight = self.request.departure.replace(
                hour=0, minute=0, second=0, microsecond=0
            )
            suboptimal_minutes = max(self.request._regional_task.suboptimalMinutes, 0)
            transit_layer = self.transport_network.transit_layer

            if (
                request._regional_task.fromLat == request._regional_task.toLat
                and request._regional_task.fromLon == request._regional_task.toLon
            ):
                lat = request._regional_task.fromLat
                lon = request._regional_task.fromLon
                transit_paths.append(
                    Trip(
                        [
                            TransitLeg(
                                transport_mode=TransportMode.TRANSIT,
                                departure_time=None,
                                distance=0.0,
                                travel_time=ZERO_SECONDS,
                                wait_time=ZERO_SECONDS,
                                geometry=shapely.LineString(((lon, lat), (lon, lat))),
                            )
                        ]
                    )
                )
            else:
                # McRapterSuboptimalPathProfileRouter needs this simple callback,
                # this could, of course, be a lambda function, but this way it’s
                # cleaner
                def list_supplier_callback(departure_time):
                    return com.conveyal.r5.profile.SuboptimalDominatingList(
                        suboptimal_minutes
                    )

                transit_router = (
                    com.conveyal.r5.profile.McRaptorSuboptimalPathProfileRouter(
                        self.transport_network,
                        request,
                        self._transit_access_times,
                        self._transit_egress_times,
                        list_supplier_callback,
                        None,
                        True,
                    )
                )
                transit_router.route()

                # `finalStatesByDepartureTime` is a hashmap of lists of router
                # states, indexed by departure times (in seconds since midnight)
                final_states = {
                    (midnight + datetime.timedelta(seconds=departure_time)): state
                    for departure_time, states in zip(
                        transit_router.finalStatesByDepartureTime.keys(),
                        transit_router.finalStatesByDepartureTime.values(),
                    )
                    for state in list(states)  # some departure times yield no results
                }

                # keep another cache layer of shortest access and egress legs
                access_legs_by_stop = {}
                egress_legs_by_stop = {}

                for departure_time, state in final_states.items():
                    trip = Trip()
                    while state:
                        if state.stop == -1:  # EgressLeg
                            try:
                                leg = egress_legs_by_stop[state.back.stop]
                            except KeyError:
                                leg = min(
                                    [
                                        self._transit_egress_paths[transport_mode][
                                            state.back.stop
                                        ]
                                        for transport_mode in self._transit_egress_paths.keys()
                                    ]
                                )
                                egress_legs_by_stop[state.back.stop] = leg
                            leg.wait_time = ZERO_SECONDS
                            leg.departure_time = (
                                midnight
                                + datetime.timedelta(seconds=state.back.time)
                                + ONE_MINUTE
                            )
                            leg.arrival_time = leg.departure_time + leg.travel_time

                        elif state.back is None:  # AccessLeg
                            try:
                                leg = access_legs_by_stop[state.stop]
                            except KeyError:
                                leg = min(
                                    [
                                        self._transit_access_paths[transport_mode][
                                            state.stop
                                        ]
                                        for transport_mode in self._transit_access_paths.keys()
                                    ]
                                )
                                access_legs_by_stop[state.stop] = leg
                            leg.wait_time = ZERO_SECONDS
                            leg.arrival_time = midnight + datetime.timedelta(
                                seconds=state.time
                            )
                            leg.departure_time = leg.arrival_time - leg.travel_time

                        else:
                            if state.pattern == -1:  # TransferLeg
                                departure_stop = state.back.stop
                                arrival_stop = state.stop

                                leg = self._transit_transfer_path(
                                    departure_stop, arrival_stop
                                )

                                leg.departure_time = (
                                    midnight
                                    + datetime.timedelta(seconds=state.back.time)
                                    + ONE_MINUTE
                                )
                                leg.arrival_time = leg.departure_time + leg.travel_time
                                leg.wait_time = (
                                    datetime.timedelta(
                                        seconds=(state.time - state.back.time)
                                    )
                                    - leg.travel_time
                                    + ONE_MINUTE  # the slack added above
                                )

                            else:  # TransitLeg
                                pattern = transit_layer.trip_patterns[state.pattern]

                                # Use the indices to look up the stop ids, which are scoped by the GTFS feed supplied
                                start_stop_id = transit_layer.get_stop_id_from_index(
                                    state.back.stop
                                ).split(":")[1]
                                end_stop = transit_layer.get_stop_id_from_index(
                                    state.stop
                                )
                                end_stop_id = end_stop.split(":")[1]
                                feed = end_stop.split(":")[0]

                                route = transit_layer.routes[pattern.routeIndex]
                                transport_mode = TransportMode(
                                    com.conveyal.r5.transit.TransitLayer.getTransitModes(
                                        route.route_type
                                    ).toString()
                                )
                                departure_time = midnight + datetime.timedelta(
                                    seconds=state.boardTime
                                )
                                travel_time = datetime.timedelta(
                                    seconds=(state.time - state.boardTime)
                                )
                                wait_time = datetime.timedelta(
                                    seconds=(state.boardTime - state.back.time)
                                )

                                # ‘hops’ in R5 terminology are the LineStrings
                                # between each pair of consecutive stops of a route
                                hops = list(pattern.getHopGeometries(transit_layer))

                                # select only the ‘hops’ between our stops, and merge
                                # them into one LineString
                                hops = hops[
                                    state.boardStopPosition : state.alightStopPosition
                                ]
                                geometry = shapely.line_merge(
                                    shapely.MultiLineString(
                                        [
                                            shapely.from_wkt(str(geometry.toText()))
                                            for geometry in hops
                                        ]
                                    )
                                )

                                # distance: based on the geometry, which might
                                # be inaccurate.

                                # We do not compute distance values if the
                                # geometry is straight lines between stops
                                # the user can still do that themselves from the
                                # inaccurate geometries: then they at least know
                                # what they committed to.
                                # TODO: add to documentation
                                if ACCURATE_GEOMETRIES:
                                    distance = shapely.ops.transform(
                                        self._crs_transformer_function,
                                        geometry,
                                    ).length
                                else:
                                    distance = None

                                leg = TransitLeg(
                                    transport_mode=transport_mode,
                                    departure_time=departure_time,
                                    distance=distance,
                                    travel_time=travel_time,
                                    wait_time=wait_time,
                                    feed=str(feed),
                                    agency_id=str(route.agency_id),
                                    route_id=str(route.route_id),
                                    start_stop_id=str(start_stop_id),
                                    end_stop_id=str(end_stop_id),
                                    geometry=geometry,
                                )

                        # we traverse in reverse order:
                        # add leg to beginning of trip,
                        # then fetch previous state (=leg)
                        trip = leg + trip
                        state = state.back

                    # R5 sometimes reports the same path more than once, skip duplicates
                    if trip not in transit_paths:
                        transit_paths.append(trip)

        return transit_paths

    @functools.cached_property
    def _transit_access_paths(self):
        access_paths = {}

        request = copy.copy(self.request)
        request._regional_task.reverseSearch = False

        street_router = com.conveyal.r5.streets.StreetRouter(
            self.transport_network.street_layer
        )
        street_router.profileRequest = request
        street_router.setOrigin(
            self.request._regional_task.fromLat,
            self.request._regional_task.fromLon,
        )
        street_router.transitStopSearch = True
        street_router.timeLimitSeconds = round(self.MAX_ACCESS_TIME.total_seconds())

        transit_layer = self.transport_network.transit_layer

        for transport_mode in request.access_modes:
            access_paths[transport_mode] = {}

            street_router.streetMode = transport_mode
            street_router.route()
            reached_stops = street_router.getReachedStops()

            for stop in reached_stops.keys():
                router_state = street_router.getStateAtVertex(
                    transit_layer.get_street_vertex_for_stop(stop)
                )
                street_segment = self._street_segment_from_router_state(
                    router_state,
                    transport_mode,
                )
                access_paths[transport_mode][stop] = AccessLeg(
                    transport_mode, street_segment
                )
        return access_paths

    @functools.cached_property
    def _transit_access_times(self):
        """
        Times to reached stops.

        In the format required by McRaptorSuboptimalPathProfileRouter.
        """
        access_times = jpype.JObject(
            {
                com.conveyal.r5.api.util.LegMode
                @ mode: gnu.trove.map.hash.TIntIntHashMap(
                    [stop for stop in reached_stops.keys()],
                    [
                        round(transfer_leg.travel_time.total_seconds())
                        for transfer_leg in reached_stops.values()
                    ],
                )
                for mode, reached_stops in self._transit_access_paths.items()
            },
            "java.util.Map<com.conveyal.r5.LegMode, gnu.trove.map.TIntIntMap>",
        )
        return access_times

    @functools.cached_property
    def _transit_egress_paths(self):
        egress_paths = {}

        request = copy.copy(self.request)
        request._regional_task.reverseSearch = True

        street_router = com.conveyal.r5.streets.StreetRouter(
            self.transport_network.street_layer
        )
        street_router.profileRequest = request
        street_router.setOrigin(
            self.request._regional_task.toLat,
            self.request._regional_task.toLon,
        )
        street_router.transitStopSearch = True
        street_router.timeLimitSeconds = round(self.MAX_EGRESS_TIME.total_seconds())

        transit_layer = self.transport_network.transit_layer

        for transport_mode in request.egress_modes:
            egress_paths[transport_mode] = {}

            street_router.streetMode = transport_mode

            street_router.route()
            reached_stops = street_router.getReachedStops()

            for stop in reached_stops.keys():
                router_state = street_router.getStateAtVertex(
                    transit_layer.get_street_vertex_for_stop(stop)
                )
                street_segment = self._street_segment_from_router_state(
                    router_state,
                    transport_mode,
                )
                egress_paths[transport_mode][stop] = EgressLeg(
                    transport_mode, street_segment
                )
        return egress_paths

    @functools.cached_property
    def _transit_egress_times(self):
        """
        Times to reached stops.

        In the format required by McRaptorSuboptimalPathProfileRouter.
        """
        egress_times = jpype.JObject(
            {
                com.conveyal.r5.api.util.LegMode
                @ mode: gnu.trove.map.hash.TIntIntHashMap(
                    [stop for stop in reached_stops.keys()],
                    [
                        round(transfer_leg.travel_time.total_seconds())
                        for transfer_leg in reached_stops.values()
                    ],
                )
                for mode, reached_stops in self._transit_egress_paths.items()
            },
            "java.util.Map<com.conveyal.r5.LegMode, gnu.trove.map.TIntIntMap>",
        )
        return egress_times

    def _transit_transfer_path(self, from_stop, to_stop):
        """Find a transfer path between two transit stops."""
        self._transfer_paths = {}
        while True:
            try:
                transfer_path = self._transfer_paths[(from_stop, to_stop)]
            except KeyError:
                request = copy.copy(self.request)

                street_router = com.conveyal.r5.streets.StreetRouter(
                    self.transport_network.street_layer
                )
                street_router.profileRequest = request
                street_router.streetMode = TransportMode.WALK

                get_coordinates_for_stop = (
                    self.transport_network.transit_layer._transit_layer.getCoordinateForStopFixed
                )
                from_stop_coordinates = get_coordinates_for_stop(from_stop)
                to_stop_coordinates = get_coordinates_for_stop(to_stop)

                from_lat = from_stop_coordinates.getY() / COORDINATE_CORRECTION_FACTOR
                from_lon = from_stop_coordinates.getX() / COORDINATE_CORRECTION_FACTOR
                to_lat = to_stop_coordinates.getY() / COORDINATE_CORRECTION_FACTOR
                to_lon = to_stop_coordinates.getX() / COORDINATE_CORRECTION_FACTOR

                street_router.setOrigin(from_lat, from_lon)
                street_router.setDestination(to_lat, to_lon)

                street_router.route()

                router_state = street_router.getState(
                    street_router.getDestinationSplit()
                )
                street_segment = self._street_segment_from_router_state(
                    router_state,
                    TransportMode.WALK,
                )

                transfer_path = self._transfer_paths[(from_stop, to_stop)] = (
                    TransferLeg(TransportMode.WALK, street_segment)
                )

                return transfer_path



#!/usr/bin/env python3


"""Fixtures related to the destinations used in routing."""


import geopandas
import pytest
import shapely

from .data_directory import DATA_DIRECTORY


SNAPPED_POPULATION_GRID_POINTS = (
    DATA_DIRECTORY / "test_snapped_population_grid_centroids.geojson"
)


@pytest.fixture(scope="session")
def population_grid():
    """Load the grid point data set."""
    import r5py.sampledata.helsinki

    yield geopandas.read_file(r5py.sampledata.helsinki.population_grid)


@pytest.fixture(scope="session")
def population_grid_points(population_grid):
    """Return the grid point data set in EPSG:4326."""
    population_grid_points = population_grid.copy()
    population_grid_points.geometry = population_grid_points.geometry.to_crs(
        "EPSG:3067"
    ).centroid.to_crs("EPSG:4326")
    yield population_grid_points


@pytest.fixture(scope="session")
def population_grid_points_first_three(population_grid_points):
    """Return the first set of three grid points."""
    yield population_grid_points[0:3]


@pytest.fixture(scope="session")
def population_grid_points_second_three(population_grid_points):
    """Return the second set of three grid points."""
    yield population_grid_points[4:7]


@pytest.fixture(scope="session")
def population_grid_points_four(population_grid_points):
    """Return four grid points."""
    yield population_grid_points[10:14]


@pytest.fixture(scope="session")
def snapped_population_grid_points():
    """Return a `geopandas.GeoDataFrame` that contains grid points snapped to the street network."""
    yield geopandas.read_file(SNAPPED_POPULATION_GRID_POINTS)


@pytest.fixture
def unreachable_stops():
    """Return a list of public transport stops that cannot be reached."""
    yield [
        1294132,
        1174101,
        1452601,
    ]


@pytest.fixture
def unsnappable_points():
    """Retrieve a set of points that cannot be snapped to the sample data network."""
    yield geopandas.GeoDataFrame(
        {
            "id": [1, 2],
            "geometry": [
                shapely.Point(48.20, 16.36),  # far away from Helsinki
                shapely.Point(-0.22, -78.51),  # even further
            ],
        },
        crs="EPSG:4326",
    )



#!/usr/bin/env python3

"""Create a copy or link of an input file in a cache directory."""


import filelock
import pathlib
import shutil

from .config import Config


__all__ = ["WorkingCopy"]


class WorkingCopy(pathlib.Path):
    """Create a copy or link of an input file in a cache directory."""

    def __new__(cls, path):
        """
        Create a copy or link of an input file in a cache directory.

        This exists because R5 creates temporary files in the directory of input
        files. This can not only be annoying clutter, but also create problems
        of concurrency, performance, etc., for instance, when the data comes
        from a shared network drive or a read-only file system.

        Arguments
        ---------
        path : str or pathlib.Path
            The file to create a copy or link of in a cache directory
        """
        # try to first create a symbolic link, if that fails (e.g., on Windows),
        # copy the file to a cache directory
        path = pathlib.Path(path).absolute()
        destination = pathlib.Path(Config().CACHE_DIR / path.name).absolute()

        with filelock.FileLock(destination.parent / f"{destination.name}.lock"):
            if not destination.exists():
                try:
                    destination.symlink_to(path)
                except OSError:
                    shutil.copyfile(f"{path}", f"{destination}")
        return destination



#!/usr/bin/env python3


"""Represent one leg of a public transport trip."""


from .trip_leg import TripLeg


__all__ = ["TransitLeg"]


class TransitLeg(TripLeg):
    """Represent one leg of a public transport trip."""



#!/usr/bin/env python3


"""Fixtures related to the R5 classpath."""


import pathlib

import pytest


R5_JAR_URL = (
    "https://github.com/r5py/r5/releases/download/v7.3-r5py/r5-v7.3-r5py-all.jar"
)
R5_JAR_SHA256 = "cb1ccad370757ba229cf17f1bedc9549ff5d77fdbb44b7a3058104fe1f243f53"
R5_JAR_SHA256_INVALID = "adfadsfadsfadsfasdfasdf"
R5_JAR_SHA256_GITHUB_ERROR_MESSAGE_WHEN_POSTING = (
    "14aa2347be79c280e4d0fd3a137fb8f5bf2863261a1e48e1a122df1a52a0f453"
)


@pytest.fixture(scope="session")
def r5_jar_cached():
    """Return a cache path for the R5 jar."""
    from r5py.util.config import Config

    yield str(Config().CACHE_DIR / pathlib.Path(R5_JAR_URL).name)


@pytest.fixture
def r5_jar_cached_invalid():
    """Return an invalid cache path for the R5 jar."""
    yield "/definitely/invalid/path/to/r5.jar"


@pytest.fixture
def r5_jar_sha256():
    """Return the SHA256 hash for the R5 jar."""
    yield R5_JAR_SHA256


@pytest.fixture
def r5_jar_sha256_invalid():
    """Return an invalid SHA256 hash for the R5 jar."""
    yield R5_JAR_SHA256_INVALID


@pytest.fixture
def r5_jar_sha256_github_error_message_when_posting():
    """Return the SHA256 hash of the GitHub error message when accidently POSTing."""
    yield R5_JAR_SHA256_GITHUB_ERROR_MESSAGE_WHEN_POSTING


@pytest.fixture()
def r5_jar_url():
    """Return the URL of the R5 jar."""
    yield R5_JAR_URL



#!/usr/bin/env python3


"""Represent one trip, consisting of one or more `TripLeg`s."""


import datetime

import shapely

from .trip_leg import TripLeg


__all__ = ["Trip"]


class Trip:
    """Represent one trip, consisting of one or more `r5py.r5.TripLeg`."""

    COLUMNS = [
        "segment",
    ] + TripLeg.COLUMNS

    def __init__(self, legs=[]):
        """
        Represent one trip, consisting of one of more `r5py.r5.TripLeg`.

        Arguments
        =========
        legs : collections.abc.Iterable
            optional list of trip legs with which to initialise this trip
        """
        self.legs = legs

    def __eq__(self, other):
        """Check whether `self` and `other` are equal."""
        if isinstance(other, self.__class__):
            return self.legs == other.legs

    def __repr__(self):
        """Return a string representation of `self`."""
        legs = ", ".join([str(leg) for leg in self.legs])
        return (
            f"<{self.__class__.__name__}: "
            f"{self.distance}m, "
            f"{self.travel_time.total_seconds()}s, "
            f"{legs}>"
        )

    def as_table(self):
        """
        Return a table (list of lists) of this trip’s legs.

        Returns
        =======
        list : detailed information about this trip and its legs (segments):
            ``segment``, ``transport_mode``, ``departure_time``, ``distance``,
            ``travel_time``, ``wait_time``, ``feed``, ``agency_id``, ``route_id``,
            ``start_stop_id``, ``end_stop_id``, ``geometry``
        """
        return [[segment] + leg.as_table_row() for segment, leg in enumerate(self.legs)]

    @property
    def distance(self):
        """Overall distance of this trip in metres (float)."""
        try:
            distance = sum([leg.distance for leg in self.legs])
        except TypeError:  # distance of a leg can be None
            distance = None
        return distance

    @property
    def geometry(self):
        """Joined geometries of all legs of this trip (shapely.LineString or shapely.MultiLineString)."""
        return shapely.line_merge(
            shapely.MultiLineString([leg.geometry for leg in self.legs])
        )

    @property
    def route_ids(self):
        """The public transport route(s) used on this trip."""
        return [leg.route_id for leg in self.legs]

    @property
    def transport_modes(self):
        """The transport mode(s) used on this trip."""
        return [leg.transport_mode for leg in self.legs]

    @property
    def travel_time(self):
        """Overall travel_time of this trip (datetime.timedelta)."""
        return sum(
            [leg.travel_time for leg in self.legs], datetime.timedelta(seconds=0)
        )

    @property
    def wait_time(self):
        """Overall wait_time of this trip (datetime.timedelta)."""
        return sum([leg.wait_time for leg in self.legs], datetime.timedelta(seconds=0))



#!/usr/bin/env python3

"""Set up a JVM and import basic java classes."""

import os
import pathlib
import shutil
import sys

import jpype
import jpype.imports

from .classpath import R5_CLASSPATH
from .config import Config
from .memory_footprint import MAX_JVM_MEMORY


__all__ = ["start_jvm"]


def start_jvm():
    """
    Start a Java Virtual Machine (JVM) if none is running already.

    Takes into account the `--max-memory` and `--verbose` command
    line and configuration options.
    """
    if not jpype.isJVMStarted():
        # preload signal handling; this, among other things, prevents some of
        # the warning messages we have been seeing
        # (cf. https://stackoverflow.com/q/15790403 and
        #  https://docs.oracle.com/en/java/javase/19/vm/signal-chaining.html )
        JVM_PATH = pathlib.Path(jpype.getDefaultJVMPath()).resolve()
        if sys.platform == "linux":
            try:
                LIBJSIG = next(JVM_PATH.parent.glob("**/libjsig.so"))
                os.environ["LD_PRELOAD"] = str(LIBJSIG)
            except StopIteration:
                pass  # don’t fail completely if libjsig not found
        elif sys.platform == "darwin":
            try:
                LIBJSIG = next(JVM_PATH.parent.glob("**/libjsig.dylib"))
                os.environ["DYLD_INSERT_LIBRARIES"] = str(LIBJSIG)
            except StopIteration:
                pass  # don’t fail completely if libjsig not found

        TEMP_DIR = Config().TEMP_DIR

        jpype.startJVM(
            f"-Xmx{MAX_JVM_MEMORY:d}",
            "-Xcheck:jni",
            "-Xrs",  # https://stackoverflow.com/q/34951812
            "-Duser.language=en",  # Set a default locale, …
            "-Duser.country=US",  # … as R5 formats numeric return …
            "-Duser.variant=",  # … values as a localised string
            f"-Djava.io.tmpdir={TEMP_DIR}",
            classpath=[R5_CLASSPATH],
            interrupt=True,
        )

        # Add shutdown hook that cleans up the temporary directory
        @jpype.JImplements("java.lang.Runnable")
        class ShutdownHookToCleanUpTempDir:
            @jpype.JOverride
            def run(self):
                shutil.rmtree(TEMP_DIR)

        import java.lang

        java.lang.Runtime.getRuntime().addShutdownHook(
            java.lang.Thread(ShutdownHookToCleanUpTempDir())
        )

        if not Config().arguments.verbose:
            import ch.qos.logback.classic
            import java.io
            import java.lang
            import org.slf4j.LoggerFactory

            logger_context = org.slf4j.LoggerFactory.getILoggerFactory()
            for log_target in (
                "com.conveyal.gtfs",
                "com.conveyal.osmlib",
                "com.conveyal.r5",
                "com.conveyal.r5.profile.ExecutionTimer",
                "com.conveyal.r5.profile.FastRaptorWorker",
                "graphql.GraphQL",
                "org.eclipse.jetty",
                "org.hsqldb.persist.Logger" "org.mongodb.driver.connection",
            ):
                logger_context.getLogger(log_target).setLevel(
                    ch.qos.logback.classic.Level.valueOf("OFF")
                )

            if sys.platform == "win32":  # Windows
                null_stream = java.io.PrintStream("NUL")
            else:
                null_stream = java.io.PrintStream("/dev/null")
            java.lang.System.setErr(null_stream)
            java.lang.System.setOut(null_stream)


# The JVM should be started before we attempt to import any Java package.
# If we run `start_jvm()` before another `import` statement, linting our
# code would result in many E402 (‘Module level import not at top of file’)
# warnings, if the JVM would start implicitely when `__file__` is imported,
# we would end up with F401 (‘Module imported but unused’) warnings.

# This below is a middle way: We don’t start the JVM right away, only
# when `start_jvm()` is called. However, if we attempt to import a
# Java package (or, more precisely, a package that’s likely to be a
# Java package), the `import` statement would trigger `start_jvm()`

# see:
# https://github.com/jpype-project/jpype/blob/master/jpype/imports.py#L146


class _JImportLoaderThatStartsTheJvm(jpype.imports._JImportLoader):
    """Find Java packages for import statements, start JVM before that."""

    def find_spec(self, name, path, target=None):
        # we got this far in `sys.meta_path` (no other finder/loader
        # knew about the package we try to load), and naturally, we’re
        # towards the end of that list.

        # Let’s assume the requested packages is a Java package,
        # and start the JVM
        start_jvm()

        # then go the standard jpype way:
        return super().find_spec(name, path, target)


# replace jpype’s _JImportLoader with our own:
for i, finder in enumerate(sys.meta_path):
    if isinstance(finder, jpype.imports._JImportLoader):
        sys.meta_path[i] = _JImportLoaderThatStartsTheJvm()



#!/usr/bin/env python3


"""Assign a cluster label column to a point-geometry GeoDataFrame."""


import warnings

import geopandas
import numpy
import shapely
import sklearn.cluster

from .good_enough_equidistant_crs import GoodEnoughEquidistantCrs


__all__ = ["SpatiallyClusteredGeoDataFrame"]


class SpatiallyClusteredGeoDataFrame(geopandas.GeoDataFrame):
    """Assign a cluster label column to a point-geometry GeoDataFrame."""

    def __init__(self, data, *args, eps=200.0, min_cluster_size=3, **kwargs):
        """
        Assign a cluster label column to a point-geometry GeoDataFrame.

        Arguments:
        ----------
        data : geopandas.GeoDataFrame
            input data set
        eps : int | float
            EPS parameter to a DBSCAN cluster algorithm, the maximum
            intra-cluster distance between two points
        *args, **kwargs: passed to geopandas.GeoDataFrame.__init__()
        """
        geopandas.GeoDataFrame.__init__(self, *args, **kwargs)

        EQUIDISTANT_CRS = GoodEnoughEquidistantCrs(
            shapely.box(*data.to_crs("EPSG:4326").geometry.total_bounds)
        )

        # loosely based on:
        # https://github.com/geopandas/scipy2018-geospatial-data/blob/master/08-clustering.ipynb

        coordinates = numpy.vstack(
            data.to_crs(EQUIDISTANT_CRS)["geometry"]
            .apply(lambda geometry: numpy.hstack(geometry.xy))
            .values
        )

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                "Could not find the number of physical cores",
                category=UserWarning,
            )
            data["cluster"] = (
                sklearn.cluster.DBSCAN(
                    eps=eps,
                    min_samples=min_cluster_size,
                    n_jobs=-1,
                )
                .fit(coordinates)
                .labels_
            )

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=(
                    "You are adding a column named 'geometry' to a GeoDataFrame "
                    "constructed without an active geometry column"
                ),
                category=FutureWarning,
            )
            for column in data.columns:
                self[column] = data[column]
            self.set_geometry("geometry")



#!/usr/bin/env python3


"""Fixtures describing the expected routing results."""


import geopandas
import pandas
import pytest

from .data_directory import DATA_DIRECTORY


DETAILED_ITINERARIES_BICYCLE = (
    DATA_DIRECTORY / "test_detailed_itineraries_bicycle.gpkg.zip"
)
DETAILED_ITINERARIES_CAR = DATA_DIRECTORY / "test_detailed_itineraries_car.gpkg.zip"
DETAILED_ITINERARIES_TRANSIT = (
    DATA_DIRECTORY / "test_detailed_itineraries_transit.gpkg.zip"
)
DETAILED_ITINERARIES_WALK = DATA_DIRECTORY / "test_detailed_itineraries_walk.gpkg.zip"

ISOCHRONES_FROM_MULTIPLE_ORIGINS = (
    DATA_DIRECTORY / "test_isochrones_from_multiple_origins.gpkg.zip"
)
ISOCHRONES_BICYCLE = DATA_DIRECTORY / "test_isochrones_bicycle.gpkg.zip"
ISOCHRONES_CAR = DATA_DIRECTORY / "test_isochrones_car.gpkg.zip"
ISOCHRONES_TRANSIT = DATA_DIRECTORY / "test_isochrones_transit.gpkg.zip"
ISOCHRONES_WALK = DATA_DIRECTORY / "test_isochrones_walk.gpkg.zip"

TRAVEL_TIMES_BICYCLE = DATA_DIRECTORY / "test_travel_times_bicycle.csv"
TRAVEL_TIMES_CAR = DATA_DIRECTORY / "test_travel_times_car.csv"
TRAVEL_TIMES_TRANSIT = DATA_DIRECTORY / "test_travel_times_transit.csv"
TRAVEL_TIMES_WALK = DATA_DIRECTORY / "test_travel_times_walk.csv"

WALKING_DETAILS_NOT_SNAPPED = DATA_DIRECTORY / "test_walking_details_not_snapped.csv"
WALKING_DETAILS_SNAPPED = DATA_DIRECTORY / "test_walking_details_snapped.csv"
WALKING_TIMES_NOT_SNAPPED = DATA_DIRECTORY / "test_walking_times_not_snapped.csv"
WALKING_TIMES_SNAPPED = DATA_DIRECTORY / "test_walking_times_snapped.csv"


@pytest.fixture
def detailed_itineraries_bicycle():
    """Retrieve expected detailed itineraries for cycling."""
    yield geopandas.read_file(DETAILED_ITINERARIES_BICYCLE)


@pytest.fixture
def detailed_itineraries_car():
    """Retrieve expected detailed itineraries for driving."""
    yield geopandas.read_file(DETAILED_ITINERARIES_CAR)


@pytest.fixture
def detailed_itineraries_transit():
    """Retrieve expected detailed itineraries for public transport."""
    yield geopandas.read_file(DETAILED_ITINERARIES_TRANSIT)


@pytest.fixture
def detailed_itineraries_walk():
    """Retrieve expected detailed itineraries for walking."""
    yield geopandas.read_file(DETAILED_ITINERARIES_WALK)


@pytest.fixture
def isochrones_from_multiple_origins():
    """Retrieve expected isochrones for multiple origins."""
    yield geopandas.read_file(ISOCHRONES_FROM_MULTIPLE_ORIGINS)


@pytest.fixture
def isochrones_bicycle():
    """Retrieve expected isochrones for cycling."""
    yield geopandas.read_file(ISOCHRONES_BICYCLE)


@pytest.fixture
def isochrones_car():
    """Retrieve expected isochrones for driving."""
    yield geopandas.read_file(ISOCHRONES_CAR)


@pytest.fixture
def isochrones_transit():
    """Retrieve expected isochrones for public transport."""
    yield geopandas.read_file(ISOCHRONES_TRANSIT)


@pytest.fixture
def isochrones_walk():
    """Retrieve expected isochrones for walking."""
    yield geopandas.read_file(ISOCHRONES_WALK)


@pytest.fixture
def travel_times_bicycle():
    """Retrieve expected travel times for cycling."""
    yield pandas.read_csv(TRAVEL_TIMES_BICYCLE)


@pytest.fixture
def travel_times_car():
    """Retrieve expected travel times for driving."""
    yield pandas.read_csv(TRAVEL_TIMES_CAR)


@pytest.fixture
def travel_times_transit():
    """Retrieve expected travel times for public transport."""
    yield pandas.read_csv(TRAVEL_TIMES_TRANSIT)


@pytest.fixture
def travel_times_walk():
    """Retrieve expected travel times for walking."""
    yield pandas.read_csv(TRAVEL_TIMES_WALK)


@pytest.fixture
def walking_details_not_snapped():
    """Load expected walking details if snapping is enabled."""
    yield pandas.read_csv(WALKING_DETAILS_NOT_SNAPPED)


@pytest.fixture
def walking_details_snapped():
    """Load expected walking details if snapping is enabled."""
    yield pandas.read_csv(WALKING_DETAILS_SNAPPED)


@pytest.fixture
def walking_times_not_snapped():
    """Load expected walking times if snapping is disabled."""
    yield pandas.read_csv(WALKING_TIMES_NOT_SNAPPED)


@pytest.fixture
def walking_times_snapped():
    """Load expected walking times if snapping is enabled."""
    yield pandas.read_csv(WALKING_TIMES_SNAPPED)



#!/usr/bin/env python3


"""Represent one leg of a trip, specifically egress from a public transport stop."""


from .transfer_leg import TransferLeg


__all__ = ["EgressLeg"]


class EgressLeg(TransferLeg):
    """Represent one leg of a trip, specifically egress from a public transport stop."""



#!/usr/bin/env python3


"""Wraps a com.conveyal.r5.streets.StreetLayer."""


import functools

import jpype
import jpype.types
import shapely

from .transport_mode import TransportMode
from ..util import start_jvm

import com.conveyal.r5


__all__ = ["StreetLayer"]


start_jvm()


EMPTY_POINT = shapely.Point()


class StreetLayer:
    """Wrap a com.conveyal.r5.streets.StreetLayer."""

    @classmethod
    def from_r5_street_layer(cls, street_layer):
        """
        Create a StreetLayer from a com.conveyal.r5.streets.StreetLayer.

        Arguments
        ---------
        street_layer : com.conveyal.r5.streets.StreetLayer
        """
        instance = cls()
        instance._street_layer = street_layer
        return instance

    @functools.cached_property
    def extent(self):
        """The geographic area covered, as a `shapely.box`."""
        envelope = self._street_layer.envelope
        return shapely.box(
            envelope.getMinX(),
            envelope.getMinY(),
            envelope.getMaxX(),
            envelope.getMaxY(),
        )

    def find_split(
        self,
        point,
        radius=com.conveyal.r5.streets.StreetLayer.LINK_RADIUS_METERS,
        street_mode=TransportMode.WALK,
    ):
        """
        Find a location on an existing street near `point`.

        Arguments
        ---------
        point : shapely.Point
            Find a location close to this point
        radius : float
            Search radius around `point`
        street_mode : travel mode that the snapped-to street should allow

        Returns
        -------
        shapely.Point
            Closest location on the street network or `POINT EMPTY` if no
            such location could be found within `radius`
        """
        try:
            split = self._street_layer.findSplit(point.y, point.x, radius, street_mode)
            return shapely.Point(
                split.fixedLon / com.conveyal.r5.streets.VertexStore.FIXED_FACTOR,
                split.fixedLat / com.conveyal.r5.streets.VertexStore.FIXED_FACTOR,
            )
        except (AttributeError, TypeError):
            return EMPTY_POINT


@jpype._jcustomizer.JConversion(
    "com.conveyal.r5.streets.StreetLayer", exact=StreetLayer
)
def _cast_StreetLayer(java_class, object_):
    return object_._street_layer



#!/usr/bin/env python3

"""Configuration and fixtures for testing r5py."""

# This is a init file common to all tests. It is automatically sourced
# by pytest et al.

# This file imports all files in the conftest.d directory

# Define common constants (e.g., paths to test data) and fixtures (e.g.,
# transport network) there and import the fixtures into conftest_d/__init__.py.

from .conftest_d import *  # noqa: F401,F403



#!/usr/bin/env python3


"""TemporaryDirectory that cleans up gracefully (https://bugs.python.org/issue26660)."""


# see also https://www.scivision.dev/python-tempfile-permission-error-windows/
# (Python 3.10 has a fix, so as soon as we deprecate 3.9, this can be changed to
# `with tempfile.TemporaryDirectory(ignore_cleanup_errors=True)`


import pathlib
import tempfile


class TemporaryDirectory:
    """A `TemporaryDirectory` that cleans up gracefully (https://bugs.python.org/issue26660)."""

    def __init__(self):
        """Create a `TemporaryDirectory` that cleans up gracefully (https://bugs.python.org/issue26660)."""
        self.temp_directory = tempfile.TemporaryDirectory()

    def __enter__(self):
        """Provide a context."""
        return pathlib.Path(self.temp_directory.name)

    def __exit__(self, exception_type, exception_value, traceback):
        """Exit context."""
        try:
            self.temp_directory.cleanup()
        except (PermissionError, NotADirectoryError):
            pass  # let the operating system take care of it



#!/usr/bin/env python3


import pytest
import pytest_lazy_fixtures

import r5py.util


class TestFileDigest:
    @pytest.mark.parametrize(
        ["input_file", "expected_digest"],
        [
            (
                pytest_lazy_fixtures.lf("file_digest_test_file_as_pathlib_path"),
                pytest_lazy_fixtures.lf("file_digest_blake2s"),
            ),
            (
                pytest_lazy_fixtures.lf("file_digest_test_file_as_str"),
                pytest_lazy_fixtures.lf("file_digest_blake2s"),
            ),
        ],
    )
    def test_file_digest(self, input_file, expected_digest):
        assert r5py.util.FileDigest(input_file) == expected_digest

    @pytest.mark.parametrize(
        ["input_file", "digest_algorithm", "expected_digest"],
        [
            (
                pytest_lazy_fixtures.lf("file_digest_test_file_as_pathlib_path"),
                "blake2s",
                pytest_lazy_fixtures.lf("file_digest_blake2s"),
            ),
            (
                pytest_lazy_fixtures.lf("file_digest_test_file_as_str"),
                "blake2s",
                pytest_lazy_fixtures.lf("file_digest_blake2s"),
            ),
            (
                pytest_lazy_fixtures.lf("file_digest_test_file_as_pathlib_path"),
                "blake2b",
                pytest_lazy_fixtures.lf("file_digest_blake2b"),
            ),
            (
                pytest_lazy_fixtures.lf("file_digest_test_file_as_str"),
                "blake2b",
                pytest_lazy_fixtures.lf("file_digest_blake2b"),
            ),
            (
                pytest_lazy_fixtures.lf("file_digest_test_file_as_pathlib_path"),
                "sha256",
                pytest_lazy_fixtures.lf("file_digest_sha256"),
            ),
            (
                pytest_lazy_fixtures.lf("file_digest_test_file_as_str"),
                "sha256",
                pytest_lazy_fixtures.lf("file_digest_sha256"),
            ),
        ],
    )
    def test_file_digest_algorithms(
        self, input_file, digest_algorithm, expected_digest
    ):
        assert r5py.util.FileDigest(input_file, digest_algorithm) == expected_digest



#!/usr/bin/env python3


"""Wraps a com.conveyal.r5.transit.TransportNetwork."""


import functools
import hashlib
import pathlib
import warnings

import jpype
import jpype.types

from .street_layer import StreetLayer
from .transit_layer import TransitLayer
from .transport_mode import TransportMode
from ..util import Config, contains_gtfs_data, FileDigest, start_jvm, WorkingCopy

import com.conveyal.gtfs
import com.conveyal.osmlib
import com.conveyal.r5
import java.io


__all__ = ["TransportNetwork"]


PACKAGE = __package__.split(".")[0]


start_jvm()


class TransportNetwork:
    """Wrap a com.conveyal.r5.transit.TransportNetwork."""

    def __init__(self, osm_pbf, gtfs=[]):
        """
        Load a transport network.

        Arguments
        ---------
        osm_pbf : str | pathlib.Path
            file path of an OpenStreetMap extract in PBF format
        gtfs : str | pathlib.Path | list[str] | list[pathlib.Path]
            path(s) to public transport schedule information in GTFS format
        """
        osm_pbf = WorkingCopy(osm_pbf)
        if isinstance(gtfs, (str, pathlib.Path)):
            gtfs = [gtfs]
        gtfs = [WorkingCopy(path) for path in gtfs]

        # a hash representing all input files
        digest = hashlib.sha256(
            "".join([FileDigest(osm_pbf)] + [FileDigest(path) for path in gtfs]).encode(
                "utf-8"
            )
        ).hexdigest()

        try:
            transport_network = self._load_pickled_transport_network(
                Config().CACHE_DIR / f"{digest}.transport_network"
            )
        except FileNotFoundError:
            transport_network = com.conveyal.r5.transit.TransportNetwork()
            transport_network.scenarioId = PACKAGE

            osm_mapdb = Config().CACHE_DIR / f"{digest}.mapdb"
            osm_file = com.conveyal.osmlib.OSM(f"{osm_mapdb}")
            osm_file.intersectionDetection = True
            osm_file.readFromFile(f"{osm_pbf}")

            transport_network.streetLayer = com.conveyal.r5.streets.StreetLayer()
            transport_network.streetLayer.parentNetwork = transport_network
            transport_network.streetLayer.loadFromOsm(osm_file)
            transport_network.streetLayer.indexStreets()

            transport_network.transitLayer = com.conveyal.r5.transit.TransitLayer()
            transport_network.transitLayer.parentNetwork = transport_network
            for gtfs_file in gtfs:
                gtfs_feed = com.conveyal.gtfs.GTFSFeed.readOnlyTempFileFromGtfs(
                    f"{gtfs_file}"
                )
                transport_network.transitLayer.loadFromGtfs(gtfs_feed)
                gtfs_feed.close()

            transport_network.streetLayer.associateStops(transport_network.transitLayer)
            transport_network.streetLayer.buildEdgeLists()

            transport_network.transitLayer.rebuildTransientIndexes()

            transfer_finder = com.conveyal.r5.transit.TransferFinder(transport_network)
            transfer_finder.findTransfers()
            transfer_finder.findParkRideTransfer()

            transport_network.transitLayer.buildDistanceTables(None)

            osm_file.close()  # not needed after here?

            self._save_pickled_transport_network(
                transport_network, Config().CACHE_DIR / f"{digest}.transport_network"
            )

        self._transport_network = transport_network

    @classmethod
    def from_directory(cls, path):
        """
        Find input data in `path`, load an `r5py.TransportNetwork`.

        This mimicks r5r’s behaviour to accept a directory path
        as the only input to `setup_r5()`.

        If more than one OpenStreetMap extract (`.osm.pbf`) is
        found in `path`, the (alphabetically) first one is used.
        In case *no* OpenStreetMap extract is found, a `FileNotFound`
        exception is raised. Any and all GTFS data files are used.

        Arguments
        ---------
        path : str
            directory path in which to search for GTFS and .osm.pbf files

        Returns
        -------
        TransportNetwork
            A fully initialised r5py.TransportNetwork
        """
        path = pathlib.Path(path)
        try:
            potential_osm_pbf_files = sorted(path.glob("*.osm.pbf"))
            osm_pbf = potential_osm_pbf_files[0]
            if len(potential_osm_pbf_files) > 1:
                warnings.warn(
                    (
                        f"Found more than one OpenStreetMap extract file (`.osm.pbf`), "
                        f"using alphabetically first one ({osm_pbf.name})"
                    ),
                    RuntimeWarning,
                )
        except IndexError:
            raise FileNotFoundError(
                f"Could not find any OpenStreetMap extract file (`.osm.pbf`) in {path.absolute()}"
            )
        gtfs = [
            potential_gtfs_file
            for potential_gtfs_file in path.glob("*.zip")
            if contains_gtfs_data(potential_gtfs_file)
        ]

        return cls(osm_pbf, gtfs)

    def __enter__(self):
        """Provide a context."""
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """Exit context."""
        return False

    @property
    def extent(self):
        """The geographic area covered, as a `shapely.box`."""
        # TODO: figure out how to get the extent of the GTFS schedule,
        # then find the smaller extent of the two (or the larger one?)
        return self.street_layer.extent

    @property
    def linkage_cache(self):
        """Expose the `TransportNetwork`’s `linkageCache` to Python."""
        return self._transport_network.linkageCache

    def _load_pickled_transport_network(self, path):
        try:
            input_file = java.io.File(f"{path}")
            return com.conveyal.r5.kryo.KryoNetworkSerializer.read(input_file)
        except java.io.FileNotFoundException:
            raise FileNotFoundError

    def _save_pickled_transport_network(self, transport_network, path):
        output_file = java.io.File(f"{path}")
        com.conveyal.r5.kryo.KryoNetworkSerializer.write(transport_network, output_file)

    def snap_to_network(
        self,
        points,
        radius=com.conveyal.r5.streets.StreetLayer.LINK_RADIUS_METERS,
        street_mode=TransportMode.WALK,
    ):
        """
        Snap `points` to valid locations on the network.

        Arguments
        ---------
        points : geopandas.GeoSeries
            point geometries that will be snapped to the network
        radius : float
            Search radius around each `point`
        street_mode : travel mode that the snapped-to street should allow

        Returns
        -------
        geopandas.GeoSeries
            point geometries that have been snapped to the network,
            using the same index and order as the input `points`
        """
        return points.apply(
            functools.partial(
                self.street_layer.find_split,
                radius=radius,
                street_mode=street_mode,
            )
        )

    @functools.cached_property
    def street_layer(self):
        """Expose the `TransportNetwork`’s `streetLayer` to Python."""
        return StreetLayer.from_r5_street_layer(self._transport_network.streetLayer)

    @property
    def timezone(self):
        """Determine the timezone of the GTFS data."""
        return self._transport_network.getTimeZone()

    @functools.cached_property
    def transit_layer(self):
        """Expose the `TransportNetwork`’s `transitLayer` to Python."""
        return TransitLayer.from_r5_transit_layer(self._transport_network.transitLayer)


@jpype._jcustomizer.JConversion(
    "com.conveyal.r5.transit.TransportNetwork", exact=TransportNetwork
)
def _cast_TransportNetwork(java_class, object_):
    return object_._transport_network



#!/usr/bin/env python3

"""Convert a snack_case formated string to a camelCase format."""


__all__ = ["snake_to_camel_case"]


def snake_to_camel_case(snake_case):
    """Convert `snake_case` to camelCase spelling."""
    if "_" in snake_case:
        words = snake_case.split("_")
        words = [words[0].lower()] + [word.title() for word in words[1:]]
        camel_case = "".join(words)
    else:
        camel_case = snake_case[0].lower() + snake_case[1:]
    return camel_case



#!/usr/bin/env python3

import datetime

import geopandas
import geopandas.testing
import pandas
import pandas.testing
import pytest
import pytest_lazy_fixtures
import shapely

import r5py
import r5py.util.exceptions


class TestIsochrones:
    def test_isochrones_initialization(
        self,
        transport_network,
        origin_point,
        departure_datetime,
    ):

        isochrones = r5py.Isochrones(
            transport_network,
            origins=origin_point,
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.TRANSIT],
            isochrones=pandas.timedelta_range(
                start=datetime.timedelta(minutes=5),
                end=datetime.timedelta(hours=0.5),
                freq=datetime.timedelta(minutes=5),
            ),
        )

        assert isinstance(isochrones, r5py.Isochrones)
        assert isinstance(isochrones, geopandas.GeoDataFrame)
        assert isochrones.geometry.geom_type.unique() == ["MultiLineString"]

        assert isinstance(isochrones.origins, geopandas.GeoDataFrame)

        assert isochrones.origins.shape == origin_point.shape

    def test_isochrones_origin_shapely_point(
        self,
        transport_network,
        origin_point,
        departure_datetime,
    ):
        origin_point = origin_point.iat[0, 2]
        assert isinstance(origin_point, shapely.Point)

        _ = r5py.Isochrones(
            transport_network,
            origins=origin_point,
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.TRANSIT],
            isochrones=pandas.timedelta_range(
                start=datetime.timedelta(minutes=15),
                end=datetime.timedelta(hours=2),
                freq=datetime.timedelta(minutes=15),
            ),
        )

    def test_isochrones_from_multiple_origins(
        self,
        transport_network,
        multiple_origins,
        departure_datetime,
        isochrones_from_multiple_origins,  # expected
    ):
        isochrones = r5py.Isochrones(
            transport_network,
            origins=multiple_origins,
            departure=departure_datetime,
            isochrones=[5, 10],
        )
        isochrones["travel_time"] = isochrones["travel_time"].apply(
            lambda t: round(t.total_seconds() / 60)
        )

        geopandas.testing.assert_geodataframe_equal(
            isochrones,
            isochrones_from_multiple_origins,
            check_less_precise=True,  # geometries
            check_dtype=False,
            normalize=True,
        )

    @pytest.mark.parametrize(
        [
            "transport_mode",
            "expected_isochrones",
        ],
        [
            (
                r5py.TransportMode.BICYCLE,
                pytest_lazy_fixtures.lf("isochrones_bicycle"),
            ),
            (
                r5py.TransportMode.CAR,
                pytest_lazy_fixtures.lf("isochrones_car"),
            ),
            (
                r5py.TransportMode.TRANSIT,
                pytest_lazy_fixtures.lf("isochrones_transit"),
            ),
            (
                r5py.TransportMode.WALK,
                pytest_lazy_fixtures.lf("isochrones_walk"),
            ),
        ],
    )
    def test_isochrones_multimodal(
        self,
        transport_network,
        origin_point,
        departure_datetime,
        transport_mode,
        expected_isochrones,
    ):
        isochrones = r5py.Isochrones(
            transport_network,
            origins=origin_point,
            isochrones=[5],
            departure=departure_datetime,
            transport_modes=[transport_mode],
        )
        isochrones["travel_time"] = isochrones["travel_time"].apply(
            lambda t: round(t.total_seconds() / 60)
        )

        geopandas.testing.assert_geodataframe_equal(
            isochrones,
            expected_isochrones,
            check_less_precise=True,  # geometries
            check_dtype=False,
            normalize=True,
        )

    @pytest.mark.parametrize(
        [
            "requested_isochrones",
            "expected_isochrones",
        ],
        [
            (
                [15, 30],
                pandas.TimedeltaIndex(
                    [datetime.timedelta(minutes=15), datetime.timedelta(minutes=30)]
                ),
            ),
            (
                [0, 5, 10],
                pandas.TimedeltaIndex(
                    [datetime.timedelta(minutes=5), datetime.timedelta(minutes=10)]
                ),
            ),
        ],
    )
    def test_isochrones_integer_isochrones(
        self,
        transport_network,
        origin_point,
        departure_datetime,
        requested_isochrones,
        expected_isochrones,
    ):
        isochrones = r5py.Isochrones(
            transport_network,
            origins=origin_point.iat[0, 2],
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.TRANSIT],
            isochrones=requested_isochrones,
        )
        pandas.testing.assert_index_equal(isochrones.isochrones, expected_isochrones)

    def test_isochrones_unset_properties(
        self, transport_network, origin_point, departure_datetime
    ):
        isochrones = r5py.Isochrones(
            transport_network,
            origins=origin_point.iat[0, 2],
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.TRANSIT],
        )
        del isochrones._isochrones
        with pytest.raises(AttributeError):
            _ = isochrones.isochrones

    def test_isochrones_custom_percentiles(
        self,
        transport_network,
        origin_point,
        departure_datetime,
    ):
        _ = r5py.Isochrones(
            transport_network,
            origins=origin_point.iat[0, 2],
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.TRANSIT],
            percentiles=[1],
        )

    @pytest.mark.parametrize(
        [
            "point_grid_resolution",
            "point_grid_sample_ratio",
            "expected_number_of_destinations",
        ],
        [
            (100, 0.8, 2654),
            (50, 1.0, 12101),
            (200, 0.3, 285),
        ],
    )
    def test_isochrones_point_grid_parameters(
        self,
        transport_network,
        origin_point,
        departure_datetime,
        point_grid_resolution,
        point_grid_sample_ratio,
        expected_number_of_destinations,
    ):
        isochrones = r5py.Isochrones(
            transport_network,
            origins=origin_point,
            departure=departure_datetime,
            point_grid_resolution=point_grid_resolution,
            point_grid_sample_ratio=point_grid_sample_ratio,
        )

        assert len(isochrones.destinations) == pytest.approx(
            expected_number_of_destinations, abs=5
        )



#!/usr/bin/env python3


import pytest

import datetime
import geopandas
import pytest_lazy_fixtures
import shapely.geometry

import r5py
from r5py.util import start_jvm

import com.conveyal.r5
import java.time


start_jvm()


class TestRegionalTask:
    @pytest.mark.parametrize(
        ["access_modes", "expected"],
        [
            (
                [r5py.TransportMode.WALK],
                set([r5py.TransportMode.WALK]),
            ),
            (
                [r5py.TransportMode.BICYCLE],
                set([r5py.TransportMode.BICYCLE]),
            ),
            (
                [r5py.TransportMode.CAR],
                set([r5py.TransportMode.CAR]),
            ),
            (
                [r5py.TransportMode.WALK, r5py.TransportMode.BICYCLE],
                set([r5py.TransportMode.WALK, r5py.TransportMode.BICYCLE]),
            ),
            (
                [
                    r5py.TransportMode.WALK,
                    r5py.TransportMode.BICYCLE,
                    r5py.TransportMode.CAR,
                ],
                set(
                    [
                        r5py.TransportMode.WALK,
                        r5py.TransportMode.BICYCLE,
                        r5py.TransportMode.CAR,
                    ]
                ),
            ),
        ],
    )
    def test_access_mode_setter(self, regional_task, access_modes, expected):
        regional_task.access_modes = access_modes
        assert regional_task.access_modes == expected
        assert regional_task._regional_task.accessModes == r5py.RegionalTask._enum_set(
            expected, com.conveyal.r5.api.util.LegMode
        )

    @pytest.mark.parametrize(
        ["access_modes", "expected"],
        [
            (
                ["WALK"],
                set([r5py.TransportMode.WALK]),
            ),
            (
                ["BICYCLE"],
                set([r5py.TransportMode.BICYCLE]),
            ),
            (
                ["CAR"],
                set([r5py.TransportMode.CAR]),
            ),
            (
                ["WALK", "BICYCLE"],
                set([r5py.TransportMode.WALK, r5py.TransportMode.BICYCLE]),
            ),
            (
                ["WALK", "BICYCLE", "CAR"],
                set(
                    [
                        r5py.TransportMode.WALK,
                        r5py.TransportMode.BICYCLE,
                        r5py.TransportMode.CAR,
                    ]
                ),
            ),
        ],
    )
    def test_access_mode_setter_with_strings(
        self, regional_task, access_modes, expected
    ):
        regional_task.access_modes = access_modes
        assert regional_task.access_modes == expected
        assert regional_task._regional_task.accessModes == r5py.RegionalTask._enum_set(
            expected, com.conveyal.r5.api.util.LegMode
        )

    @pytest.mark.parametrize(
        [
            "departure",
            "expected",
            "expected_java_date",
            "expected_java_from_time",
        ],
        [
            (
                datetime.datetime(2022, 2, 22, 8, 30),
                datetime.datetime(2022, 2, 22, 8, 30),
                java.time.LocalDate.of(2022, 2, 22),
                30600,
            )
        ],
    )
    def test_departure_setter(
        self,
        regional_task,
        departure,
        expected,
        expected_java_date,
        expected_java_from_time,
    ):
        regional_task.departure = departure
        assert regional_task.departure == expected
        assert regional_task._regional_task.date == expected_java_date
        assert regional_task._regional_task.fromTime == expected_java_from_time

    @pytest.mark.parametrize(
        ["egress_modes", "expected"],
        [
            (
                [r5py.TransportMode.WALK],
                set([r5py.TransportMode.WALK]),
            ),
            (
                [r5py.TransportMode.BICYCLE],
                set([r5py.TransportMode.BICYCLE]),
            ),
            (
                [r5py.TransportMode.CAR],
                set([r5py.TransportMode.CAR]),
            ),
            (
                [r5py.TransportMode.WALK, r5py.TransportMode.BICYCLE],
                set([r5py.TransportMode.WALK, r5py.TransportMode.BICYCLE]),
            ),
            (
                [
                    r5py.TransportMode.WALK,
                    r5py.TransportMode.BICYCLE,
                    r5py.TransportMode.CAR,
                ],
                set(
                    [
                        r5py.TransportMode.WALK,
                        r5py.TransportMode.BICYCLE,
                        r5py.TransportMode.CAR,
                    ]
                ),
            ),
        ],
    )
    def test_egress_mode_setter(self, regional_task, egress_modes, expected):
        regional_task.egress_modes = egress_modes
        assert regional_task.egress_modes == expected
        assert regional_task._regional_task.egressModes == r5py.RegionalTask._enum_set(
            expected, com.conveyal.r5.api.util.LegMode
        )

    @pytest.mark.parametrize(
        ["egress_modes", "expected"],
        [
            (
                ["WALK"],
                set([r5py.TransportMode.WALK]),
            ),
            (
                ["BICYCLE"],
                set([r5py.TransportMode.BICYCLE]),
            ),
            (
                ["CAR"],
                set([r5py.TransportMode.CAR]),
            ),
            (
                ["WALK", "BICYCLE"],
                set([r5py.TransportMode.WALK, r5py.TransportMode.BICYCLE]),
            ),
            (
                ["WALK", "BICYCLE", "CAR"],
                set(
                    [
                        r5py.TransportMode.WALK,
                        r5py.TransportMode.BICYCLE,
                        r5py.TransportMode.CAR,
                    ]
                ),
            ),
        ],
    )
    def test_egress_mode_setter_with_strings(
        self, regional_task, egress_modes, expected
    ):
        regional_task.egress_modes = egress_modes
        assert regional_task.egress_modes == expected
        assert regional_task._regional_task.egressModes == r5py.RegionalTask._enum_set(
            expected, com.conveyal.r5.api.util.LegMode
        )

    @pytest.mark.parametrize(
        ["bicycle_stress", "expected"],
        [
            (1, 1),
            (2, 2),
            (3, 3),
            (4, 4),
        ],
    )
    def test_max_bicycle_traffic_stress_setter(
        self,
        regional_task,
        bicycle_stress,
        expected,
    ):
        regional_task.max_bicycle_traffic_stress = bicycle_stress
        assert regional_task.max_bicycle_traffic_stress == expected
        assert regional_task._regional_task.bikeTrafficStress == expected

    @pytest.mark.parametrize(
        ["max_pt_rides", "expected"],
        [
            (1, 1),
            (1232, 1232),
            (31, 31),
            (999, 999),
        ],
    )
    def test_max_public_transport_rides_setter(
        self,
        regional_task,
        max_pt_rides,
        expected,
    ):
        regional_task.max_public_transport_rides = max_pt_rides
        assert regional_task.max_public_transport_rides == expected
        assert regional_task._regional_task.maxRides == expected

    @pytest.mark.parametrize(
        ["max_time", "expected", "expected_java"],
        [
            (
                datetime.timedelta(hours=1),
                datetime.timedelta(hours=1),
                60,
            ),
            (
                datetime.timedelta(minutes=15),
                datetime.timedelta(minutes=15),
                15,
            ),
            (
                datetime.timedelta(hours=7, minutes=23),
                datetime.timedelta(hours=7, minutes=23),
                443,
            ),
        ],
    )
    def test_max_time_setter(
        self,
        regional_task,
        max_time,
        expected,
        expected_java,
    ):
        regional_task.max_time = max_time
        assert regional_task.max_time == expected
        assert regional_task._regional_task.streetTime == expected_java
        assert regional_task._regional_task.maxTripDurationMinutes == expected_java
        assert regional_task._regional_task.maxCarTime == expected_java

    @pytest.mark.parametrize(
        ["max_time_cycling", "expected", "expected_java"],
        [
            (
                datetime.timedelta(hours=1),
                datetime.timedelta(hours=1),
                60,
            ),
            (
                datetime.timedelta(minutes=15),
                datetime.timedelta(minutes=15),
                15,
            ),
            (
                datetime.timedelta(hours=7, minutes=23),
                datetime.timedelta(hours=7, minutes=23),
                443,
            ),
        ],
    )
    def test_max_time_cycling_setter(
        self,
        regional_task,
        max_time_cycling,
        expected,
        expected_java,
    ):
        regional_task.max_time_cycling = max_time_cycling
        assert regional_task.max_time_cycling == expected
        assert regional_task._regional_task.maxBikeTime == expected_java

    @pytest.mark.parametrize(
        ["max_time_driving", "expected", "expected_java"],
        [
            (
                datetime.timedelta(hours=1),
                datetime.timedelta(hours=1),
                60,
            ),
            (
                datetime.timedelta(minutes=15),
                datetime.timedelta(minutes=15),
                15,
            ),
            (
                datetime.timedelta(hours=7, minutes=23),
                datetime.timedelta(hours=7, minutes=23),
                443,
            ),
        ],
    )
    def test_max_time_driving_setter(
        self,
        regional_task,
        max_time_driving,
        expected,
        expected_java,
    ):
        regional_task.max_time_driving = max_time_driving
        assert regional_task.max_time_driving == expected
        assert regional_task._regional_task.maxCarTime == expected_java

    @pytest.mark.parametrize(
        ["max_time_walking", "expected", "expected_java"],
        [
            (
                datetime.timedelta(hours=1),
                datetime.timedelta(hours=1),
                60,
            ),
            (
                datetime.timedelta(minutes=15),
                datetime.timedelta(minutes=15),
                15,
            ),
            (
                datetime.timedelta(hours=7, minutes=23),
                datetime.timedelta(hours=7, minutes=23),
                443,
            ),
        ],
    )
    def test_max_time_walking_setter(
        self,
        regional_task,
        max_time_walking,
        expected,
        expected_java,
    ):
        regional_task.max_time_walking = max_time_walking
        assert regional_task.max_time_walking == expected
        assert regional_task._regional_task.maxWalkTime == expected_java

    @pytest.mark.parametrize(
        ["percentiles"],
        [
            ([],),
            ([50],),
            ([33, 66],),
            ([25, 50, 75],),
            ([20, 40, 60, 80],),
            ([16, 33, 50, 66, 83],),
        ],
    )
    def test_allowed_number_of_percentiles(self, regional_task, percentiles):
        regional_task.percentiles = percentiles

    @pytest.mark.parametrize(
        ["percentiles"],
        [
            ([10, 20, 30, 40, 50, 60, 70, 90],),
            ([i for i in range(101)],),
        ],
    )
    def test_out_of_range_percentiles(self, regional_task, percentiles):
        with pytest.raises(
            ValueError, match="Maximum number of percentiles allowed is 5"
        ):
            regional_task.percentiles = percentiles

    def test_scenario(self, regional_task):
        from r5py.r5 import Scenario

        scenario = Scenario()
        regional_task.scenario = scenario
        assert regional_task.scenario == scenario
        assert regional_task._regional_task.scenarioId == scenario.id

    @pytest.mark.parametrize(
        ["speed_cycling", "expected", "expected_java"],
        [
            (14.4, 14.4, 4.0),
            (18, 18.0, 5.0),
            (19.8, 19.8, 5.5),
            (28.8, 28.8, 8.0),
            (36, 36, 10.0),
        ],
    )
    def test_speed_cycling_setter(
        self,
        regional_task,
        speed_cycling,
        expected,
        expected_java,
    ):
        regional_task.speed_cycling = speed_cycling
        assert regional_task.speed_cycling == expected
        assert regional_task._regional_task.bikeSpeed == pytest.approx(expected_java)

    @pytest.mark.parametrize(
        ["speed_walking", "expected", "expected_java"],
        [
            (3.6, 3.6, 1.0),
            (5.4, 5.4, 1.5),
            (7.2, 7.2, 2.0),
            (9.0, 9.0, 2.5),
        ],
    )
    def test_speed_walking_setter(
        self,
        regional_task,
        speed_walking,
        expected,
        expected_java,
    ):
        regional_task.speed_walking = speed_walking
        assert regional_task.speed_walking == expected
        assert regional_task._regional_task.walkSpeed == pytest.approx(expected_java)

    @pytest.mark.parametrize(
        ["origin"],
        [
            (shapely.geometry.Point(60, 24),),
            (shapely.geometry.Point(61, 25),),
        ],
    )
    def test_origin_setter_getter(self, regional_task, origin):
        regional_task.origin = origin
        assert regional_task.origin == origin

    @pytest.mark.parametrize(
        ["destinations"],
        [
            (pytest_lazy_fixtures.lf("population_grid_points"),),
            (pytest_lazy_fixtures.lf("population_grid_points_first_three"),),
            (pytest_lazy_fixtures.lf("population_grid_points_second_three"),),
        ],
    )
    def test_destinations_setter_getter(self, regional_task, destinations):
        regional_task.destinations = destinations
        geopandas.testing.assert_geodataframe_equal(
            regional_task.destinations,
            destinations,
        )

    @pytest.mark.parametrize(
        ["transport_modes", "expected"],
        [
            (
                ["WALK"],
                set([r5py.TransportMode.WALK]),
            ),
            (
                ["TRANSIT", "WALK"],
                set([r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK]),
            ),
            (
                ["GONDOLA", "SUBWAY"],
                set([r5py.TransportMode.GONDOLA, r5py.TransportMode.SUBWAY]),
            ),
        ],
    )
    def test_transport_modes_setter_with_strings(
        self, regional_task, transport_modes, expected
    ):
        regional_task.transport_modes = transport_modes
        assert regional_task.transport_modes == expected

    # TODO: all other methods and attributes!



#!/usr/bin/env python3

import geopandas
import pandas
import datetime
import pytest
import pytest_lazy_fixtures

import r5py
import r5py.util.exceptions


class TestTravelTimeMatrixInputValidation:
    def test_departure_time_window_warning(
        self,
        transport_network,
        population_grid_points,
        origin_point,
        departure_datetime,
    ):
        with pytest.warns(
            RuntimeWarning,
            match="The provided departure time window is below 5 minutes",
        ):
            _ = r5py.TravelTimeMatrix(
                transport_network,
                origins=origin_point,
                destinations=population_grid_points,
                departure_time_window=datetime.timedelta(
                    minutes=3
                ),  # Should throw warning
                departure=departure_datetime,
                transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
            )

    @pytest.mark.parametrize(
        [
            "origins",
            "expected_error",
        ],
        [
            (
                pytest_lazy_fixtures.lf("origins_invalid_no_id"),
                r5py.util.exceptions.NoIDColumnError,
            ),
            (
                pytest_lazy_fixtures.lf("origins_invalid_duplicate_ids"),
                r5py.util.exceptions.NonUniqueIDError,
            ),
        ],
    )
    def test_origins_invalid_data(
        self,
        transport_network,
        origins,
        departure_datetime,
        expected_error,
    ):
        with pytest.raises(expected_error):
            _ = r5py.TravelTimeMatrix(
                transport_network, origins=origins, departure=departure_datetime
            )

    def test_origins_valid_data(
        self,
        transport_network,
        origins_valid_ids,
        departure_datetime,
    ):
        _ = r5py.TravelTimeMatrix(
            transport_network,
            origins=origins_valid_ids,
            departure=departure_datetime,
        )

    @pytest.mark.parametrize(
        [
            "origins",
            "destinations",
            "expected_error",
        ],
        [
            (
                pytest_lazy_fixtures.lf("origins_invalid_no_id"),
                pytest_lazy_fixtures.lf("origins_invalid_no_id"),
                r5py.util.exceptions.NoIDColumnError,
            ),
            (
                pytest_lazy_fixtures.lf("origins_invalid_duplicate_ids"),
                pytest_lazy_fixtures.lf("origins_invalid_duplicate_ids"),
                r5py.util.exceptions.NonUniqueIDError,
            ),
            (
                pytest_lazy_fixtures.lf("origins_invalid_no_id"),
                pytest_lazy_fixtures.lf("origins_invalid_duplicate_ids"),
                r5py.util.exceptions.NoIDColumnError,
            ),
            (
                pytest_lazy_fixtures.lf("origins_invalid_duplicate_ids"),
                pytest_lazy_fixtures.lf("origins_invalid_no_id"),
                r5py.util.exceptions.NonUniqueIDError,
            ),
        ],
    )
    def test_origins_and_destinations_invalid_data(
        self,
        transport_network,
        origins,
        destinations,
        departure_datetime,
        expected_error,
    ):
        with pytest.raises(expected_error):
            _ = r5py.TravelTimeMatrix(
                transport_network,
                origins=origins,
                destinations=destinations,
                departure=departure_datetime,
            )

    @pytest.mark.parametrize(
        [
            "origins",
            "destinations",
        ],
        [
            (
                pytest_lazy_fixtures.lf("origins_valid_ids"),
                pytest_lazy_fixtures.lf("origins_valid_ids"),
            )
        ],
    )
    def test_origins_and_destinations_valid_data(
        self,
        transport_network,
        origins,
        destinations,
        departure_datetime,
    ):
        _ = r5py.TravelTimeMatrix(
            transport_network,
            origins=origins,
            destinations=destinations,
            departure=departure_datetime,
        )

    def test_try_to_route_without_origins(
        self,
        transport_network,
        departure_datetime,
    ):
        with pytest.raises(ValueError, match="No routing origins defined"):
            _ = r5py.TravelTimeMatrix(
                transport_network,
                departure=departure_datetime,
            )

    def test_try_to_route_without_destinations(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
    ):
        travel_time_matrix = r5py.TravelTimeMatrix(
            transport_network,
            origins=population_grid_points[0:3],
            departure=departure_datetime,
        )
        pandas.testing.assert_frame_equal(
            travel_time_matrix.origins,
            travel_time_matrix.destinations,
        )


class TestTravelTimeMatrix:
    def test_travel_time_matrix_initialization(
        self,
        transport_network,
        population_grid_points,
        origin_point,
        departure_datetime,
    ):
        travel_time_matrix = r5py.TravelTimeMatrix(
            transport_network,
            origins=origin_point,
            destinations=population_grid_points,
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
        )
        assert isinstance(travel_time_matrix.origins, geopandas.GeoDataFrame)
        assert isinstance(travel_time_matrix.destinations, geopandas.GeoDataFrame)

        assert travel_time_matrix.origins.shape == origin_point.shape
        assert travel_time_matrix.destinations.shape == population_grid_points.shape

    def test_travel_time_matrix_initialization_with_files(
        self,
        transport_network_files_tuple,
        population_grid_points,
        origin_point,
        departure_datetime,
    ):
        travel_time_matrix = r5py.TravelTimeMatrix(
            transport_network_files_tuple,
            origins=origin_point,
            destinations=population_grid_points,
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
        )
        assert isinstance(travel_time_matrix, r5py.TravelTimeMatrix)
        assert isinstance(travel_time_matrix, pandas.DataFrame)

    def test_all_to_all(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
    ):
        travel_time_matrix = r5py.TravelTimeMatrix(
            transport_network,
            origins=population_grid_points,
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
        )

        assert isinstance(travel_time_matrix, pandas.DataFrame)
        assert travel_time_matrix.shape == (8464, 3)
        assert travel_time_matrix.columns.to_list() == [
            "from_id",
            "to_id",
            "travel_time",
        ]
        assert travel_time_matrix["from_id"].min() == travel_time_matrix["to_id"].min()
        assert travel_time_matrix["from_id"].max() == travel_time_matrix["to_id"].max()
        assert travel_time_matrix["travel_time"].min() >= 0
        # There can be a bit of fluctuation in the maximum travel time
        assert travel_time_matrix["travel_time"].max() == pytest.approx(50, abs=3)

    def test_one_to_all(
        self,
        transport_network,
        population_grid_points,
        origin_point,
        departure_datetime,
    ):
        travel_time_matrix = r5py.TravelTimeMatrix(
            transport_network,
            origins=origin_point,
            destinations=population_grid_points,
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
        )
        assert travel_time_matrix.shape == (92, 3)
        assert travel_time_matrix["from_id"].unique() == [0]
        assert travel_time_matrix["to_id"].min() == 0
        assert travel_time_matrix["to_id"].max() == 91
        # There can be a bit of fluctuation in the maximum travel time
        assert travel_time_matrix["travel_time"].max() == pytest.approx(30, abs=3)

    def test_one_to_all_with_percentiles(
        self,
        transport_network,
        population_grid_points,
        origin_point,
        departure_datetime,
    ):
        travel_time_matrix = r5py.TravelTimeMatrix(
            transport_network,
            origins=origin_point,
            destinations=population_grid_points,
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
            percentiles=[25, 50, 75],
        )
        assert travel_time_matrix.shape == (92, 5)
        required_cols = ["travel_time_p25", "travel_time_p50", "travel_time_p75"]
        for col in required_cols:
            assert col in travel_time_matrix.columns

        # 75 percentile should always be higher or equal to 25 percentile
        check = (
            travel_time_matrix["travel_time_p75"]
            >= travel_time_matrix["travel_time_p25"]
        )
        assert False not in check.to_list()

    def test_gtfs_date_range_warnings(
        self,
        transport_network,
        population_grid_points,
        origin_point,
        departure_datetime,
    ):
        with pytest.warns(RuntimeWarning, match="Departure time"):
            _ = r5py.TravelTimeMatrix(
                transport_network,
                origins=origin_point,
                destinations=population_grid_points,
                departure=datetime.datetime(2021, 2, 22, 8, 30),  # not in GTFS data set
                transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
            )

    def test_gtfs_date_range_warnings_without_gtfs_file(
        self,
        transport_network_from_test_files_without_gtfs,
        population_grid_points,
        origin_point,
        departure_datetime,
    ):
        with pytest.warns(RuntimeWarning, match="Departure"):
            _ = r5py.TravelTimeMatrix(
                transport_network_from_test_files_without_gtfs,
                origins=origin_point,
                destinations=population_grid_points,
                departure=departure_datetime,
                transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
            )

    @pytest.mark.parametrize(
        [
            "snap_to_network",
            "expected_snap_to_network",
        ],
        [
            (
                True,
                True,
            ),
            (
                False,
                False,
            ),
        ],
    )
    def test_snap_to_network_parameter(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
        snap_to_network,
        expected_snap_to_network,
    ):
        travel_time_matrix = r5py.TravelTimeMatrix(
            transport_network,
            population_grid_points,
            departure=departure_datetime,
            snap_to_network=snap_to_network,
        )
        assert travel_time_matrix.snap_to_network == expected_snap_to_network

    @pytest.mark.parametrize(
        [
            "snap_to_network",
            "expected_travel_times",
        ],
        [
            (
                True,
                pytest_lazy_fixtures.lf("walking_times_snapped"),
            ),
            (
                False,
                pytest_lazy_fixtures.lf("walking_times_not_snapped"),
            ),
        ],
    )
    def test_snap_to_network(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
        snap_to_network,
        expected_travel_times,
    ):
        travel_times = r5py.TravelTimeMatrix(
            transport_network,
            origins=population_grid_points,
            departure=departure_datetime,
            snap_to_network=snap_to_network,
            transport_modes=[r5py.TransportMode.WALK],
        )

        travel_times = travel_times.set_index(["from_id", "to_id"]).sort_index()
        expected_travel_times = expected_travel_times.set_index(
            ["from_id", "to_id"]
        ).sort_index()

        pandas.testing.assert_frame_equal(travel_times, expected_travel_times)

    def test_snap_to_network_with_unsnappable_origins(
        self,
        transport_network,
        unsnappable_points,
        population_grid_points,
        departure_datetime,
    ):
        origins = pandas.concat(
            [population_grid_points[-3:], unsnappable_points]
        ).reset_index(drop=False)
        with pytest.warns(
            RuntimeWarning,
            match=(
                "Some (origin|destination) points could not "
                "be snapped to the street network"
            ),
        ):
            _ = r5py.TravelTimeMatrix(
                transport_network,
                origins,
                departure=departure_datetime,
                snap_to_network=True,
                transport_modes=[r5py.TransportMode.WALK],
            )

    def test_snap_to_network_with_only_unsnappable_origins(
        self,
        transport_network,
        unsnappable_points,
        departure_datetime,
    ):
        with (
            pytest.raises(
                ValueError, match="After snapping, no valid origin points remain"
            ),
            pytest.warns(
                RuntimeWarning,
                match="Some origin points could not be snapped to the street network",
            ),
        ):
            _ = r5py.TravelTimeMatrix(
                transport_network,
                unsnappable_points,
                departure=departure_datetime,
                snap_to_network=True,
                transport_modes=[r5py.TransportMode.WALK],
            )

    def test_snap_to_network_with_unsnappable_destinations(
        self,
        transport_network,
        population_grid_points,
        unsnappable_points,
        departure_datetime,
    ):
        destinations = pandas.concat(
            [population_grid_points[-3:], unsnappable_points]
        ).reset_index(drop=False)
        with pytest.warns(
            RuntimeWarning,
            match="Some destination points could not be snapped to the street network",
        ):
            _ = r5py.TravelTimeMatrix(
                transport_network,
                origins=population_grid_points,
                destinations=destinations,
                departure=departure_datetime,
                snap_to_network=True,
                transport_modes=[r5py.TransportMode.WALK],
            )

    def test_snap_to_network_with_only_unsnappable_destinations(
        self,
        transport_network,
        population_grid_points,
        unsnappable_points,
        departure_datetime,
    ):
        with (
            pytest.raises(
                ValueError, match="After snapping, no valid destination points remain"
            ),
            pytest.warns(
                RuntimeWarning,
                match="Some destination points could not be snapped to the street network",
            ),
        ):
            _ = r5py.TravelTimeMatrix(
                transport_network,
                origins=population_grid_points,
                destinations=unsnappable_points,
                departure=departure_datetime,
                snap_to_network=True,
                transport_modes=[r5py.TransportMode.WALK],
            )

    @pytest.mark.parametrize("snap_to_network", [True, False])
    def test_travel_time_between_identical_from_and_to_ids(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
        snap_to_network,
    ):
        travel_time_matrix = r5py.TravelTimeMatrix(
            transport_network,
            origins=population_grid_points,
            transport_modes=[r5py.TransportMode.WALK],
            departure=departure_datetime,
            snap_to_network=snap_to_network,
        )

        assert (
            travel_time_matrix[
                travel_time_matrix["from_id"] == travel_time_matrix["to_id"]
            ].travel_time.max()
            == 0
        )


class TestTravelTimeMatrixComputer:
    def test_travel_time_matrix_warning_and_legacy_interface(
        self,
        transport_network,
        population_grid_points,
        origin_point,
        departure_datetime,
    ):
        ttm_new = r5py.TravelTimeMatrix(
            transport_network,
            origins=origin_point,
            destinations=population_grid_points,
            departure=departure_datetime,
            transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
        )

        with pytest.warns(
            DeprecationWarning,
            match="Use `TravelTimeMatrix` instead, `TravelTimeMatrixComputer will be deprecated in a future release.",
        ):
            ttm_old = r5py.TravelTimeMatrixComputer(
                transport_network,
                origins=origin_point,
                destinations=population_grid_points,
                departure=departure_datetime,
                transport_modes=[r5py.TransportMode.TRANSIT, r5py.TransportMode.WALK],
            ).compute_travel_times()

        geopandas.testing.assert_geodataframe_equal(ttm_new, ttm_old)

    @pytest.mark.parametrize(
        [
            "transport_mode",
            "expected_travel_times",
        ],
        [
            (
                r5py.TransportMode.BICYCLE,
                pytest_lazy_fixtures.lf("travel_times_bicycle"),
            ),
            (
                r5py.TransportMode.CAR,
                pytest_lazy_fixtures.lf("travel_times_car"),
            ),
            (
                r5py.TransportMode.TRANSIT,
                pytest_lazy_fixtures.lf("travel_times_transit"),
            ),
            (
                r5py.TransportMode.WALK,
                pytest_lazy_fixtures.lf("travel_times_walk"),
            ),
        ],
    )
    def test_travel_times_by_transport_mode(
        self,
        transport_network,
        population_grid_points,
        departure_datetime,
        transport_mode,
        expected_travel_times,
    ):
        # subset to keep test comparison data sets small
        origins = population_grid_points[::5].copy()
        travel_times = r5py.TravelTimeMatrix(
            transport_network,
            origins=origins,
            departure=departure_datetime,
            transport_modes=[transport_mode],
        )

        travel_times = travel_times.reset_index(drop=True)

        pandas.testing.assert_frame_equal(
            travel_times,
            expected_travel_times,
        )



#!/usr/bin/env python3


"""Fixtures related to transport networks."""


import pytest


@pytest.fixture
def not_a_gtfs_file():
    """Return a file path of something that is not a GTFS file."""
    import r5py.sampledata.helsinki

    yield r5py.sampledata.helsinki.osm_pbf


@pytest.fixture
def gtfs_file_path():
    """Return the file path of a GTFS sample data set."""
    import r5py.sampledata.helsinki

    yield r5py.sampledata.helsinki.gtfs


@pytest.fixture
def gtfs_timezone_helsinki():
    """Return the timezone for Helsinki."""
    yield "Europe/Helsinki"


@pytest.fixture
def osm_pbf_file_path():
    """Return the path of the OSM sample data set."""
    import r5py.sampledata.helsinki

    yield r5py.sampledata.helsinki.osm_pbf


@pytest.fixture
def transport_network_files_tuple():
    """Return a tuple of transport network input test data file paths."""
    import r5py.sampledata.helsinki

    yield r5py.sampledata.helsinki.osm_pbf, [r5py.sampledata.helsinki.gtfs]


@pytest.fixture
def transport_network(transport_network_from_test_files):
    """Return an `r5py.TransportNetwork`."""
    yield transport_network_from_test_files


@pytest.fixture
def transport_network_from_test_directory():
    """Return an `r5py.TransportNetwork` initiated from a directory path."""
    import r5py
    import r5py.sampledata.helsinki
    from ..temporary_directory import TemporaryDirectory

    with TemporaryDirectory() as temp_directory:
        (temp_directory / r5py.sampledata.helsinki.osm_pbf.name).symlink_to(
            r5py.sampledata.helsinki.osm_pbf
        )
        (temp_directory / r5py.sampledata.helsinki.gtfs.name).symlink_to(
            r5py.sampledata.helsinki.gtfs
        )

        transport_network = r5py.TransportNetwork.from_directory(temp_directory)
        yield transport_network

        del transport_network


@pytest.fixture
def transport_network_from_test_files():
    """Return an `r5py.TransportNetwork` initiated from test file paths."""
    import r5py
    import r5py.sampledata.helsinki

    transport_network = r5py.TransportNetwork(
        r5py.sampledata.helsinki.osm_pbf, [r5py.sampledata.helsinki.gtfs]
    )
    yield transport_network

    del transport_network


@pytest.fixture
def transport_network_from_test_files_without_gtfs():
    """Return an `r5py.TransportNetwork` initiated without GTFS data."""
    import r5py
    import r5py.sampledata.helsinki

    transport_network = r5py.TransportNetwork(r5py.sampledata.helsinki.osm_pbf, [])
    yield transport_network

    del transport_network



#!/usr/bin/env python3

"""Utility functions, e.g., starting a JVM, and accessing configuration."""

from . import environment  # noqa: F401

from .camel_to_snake_case import camel_to_snake_case
from .config import Config
from .contains_gtfs_data import contains_gtfs_data
from .data_validation import check_od_data_set
from .file_digest import FileDigest
from .good_enough_equidistant_crs import GoodEnoughEquidistantCrs
from .jvm import start_jvm
from .parse_int_date import parse_int_date
from .snake_to_camel_case import snake_to_camel_case
from .spatially_clustered_geodataframe import SpatiallyClusteredGeoDataFrame
from .working_copy import WorkingCopy

__all__ = [
    "camel_to_snake_case",
    "check_od_data_set",
    "Config",
    "contains_gtfs_data",
    "FileDigest",
    "GoodEnoughEquidistantCrs",
    "parse_int_date",
    "snake_to_camel_case",
    "SpatiallyClusteredGeoDataFrame",
    "start_jvm",
    "WorkingCopy",
]



#!/usr/bin/env python3

"""Utility functions for validating input data."""


from .exceptions import NoIDColumnError, NonUniqueIDError, NoCrsError


__all__ = ["check_od_data_set"]


def check_od_data_set(od_data_set):
    """Check whether an origin/destination data set fulfils certain minimum criteria.

    Checks whether `od_data_set` has an `id` column with unique values, and a coordinate
    reference system defined. Raises a `r5py.util.exceptions.NoIDColumnError`, a
    `r5py.util.exceptions.NonUniqueIDError`, or a `r5py.util.exceptions.NoCrsError`
    otherwise.

    Arguments
    ---------
        od_data_set : geopandas.GeoDataFrame
            The origin/destination data set to check.
    """
    if "id" not in od_data_set.columns:
        raise NoIDColumnError("Data set must contain an 'id' column.")
    if not od_data_set.id.is_unique:
        raise NonUniqueIDError("Id values must be unique.")
    if od_data_set.crs is None:
        raise NoCrsError("Data set has to have a coordinate reference system defined.")



#!/usr/bin/env python3


import datetime
import importlib
import pathlib
import os
import sys

import r5py.util.config
from r5py.util.config import CACHE_MAX_AGE


class TestConfig:
    def test_setdefaulthome(self):
        original_home = os.environ["HOME"]
        del os.environ["HOME"]
        importlib.reload(r5py.util.config)
        assert os.environ["HOME"] == "."
        os.environ["HOME"] = original_home

    def test_custom_temporary_directory(self, tmp_path):
        sys.argv.extend(
            [
                "--temporary-directory",
                f"{tmp_path}",
            ]
        )
        assert r5py.util.config.Config().arguments.temporary_directory.is_relative_to(
            tmp_path
        )
        sys.argv = sys.argv[:-2]

    def test_cache_clearing(self):
        config = r5py.util.config.Config()

        past_best_by_date = (
            datetime.datetime.now() - CACHE_MAX_AGE - datetime.timedelta(seconds=1)
        ).timestamp()

        expired_file = pathlib.Path(config.CACHE_DIR / "expired-file")
        expired_file.touch()
        os.utime(expired_file, (past_best_by_date, past_best_by_date))

        config.__dict__.pop("CACHE_DIR", None)  # clear functools.cached_property
        _ = config.CACHE_DIR  # re-evaluate cache dir contents

        assert not expired_file.exists()



#!/usr/bin/env python3


import hashlib

import pytest

from r5py.util.validating_requests_session import (
    ChecksumFailed,
    ValidatingRequestsSession,
)


class TestValidatingRequestSession:
    def test_initialisation(self):
        validating_request_session = ValidatingRequestsSession()
        assert validating_request_session._algorithm == hashlib.sha256

    def test_get(self, r5_jar_url, r5_jar_sha256):
        with (
            ValidatingRequestsSession() as session,
            session.get(r5_jar_url, r5_jar_sha256) as response,
        ):
            assert response.content

    def test_get_invalid_checksum(self, r5_jar_url, r5_jar_sha256_invalid):
        with pytest.raises(ChecksumFailed):
            with (
                ValidatingRequestsSession() as session,
                session.get(r5_jar_url, r5_jar_sha256_invalid) as response,
            ):
                assert response.content

    def test_post(self, r5_jar_url, r5_jar_sha256_github_error_message_when_posting):
        with (
            ValidatingRequestsSession() as session,
            session.post(
                r5_jar_url, r5_jar_sha256_github_error_message_when_posting
            ) as response,
        ):
            assert response.content

    def test_post_invalid_checksum(self, r5_jar_url, r5_jar_sha256_invalid):
        with pytest.raises(ChecksumFailed):
            with (
                ValidatingRequestsSession() as session,
                session.post(r5_jar_url, r5_jar_sha256_invalid) as response,
            ):
                assert response.content



#!/usr/bin/env python3


"""Fixtures related to routing origins."""


import warnings

import geopandas
import pytest

from .data_directory import DATA_DIRECTORY


ORIGINS_INVALID_NO_ID = DATA_DIRECTORY / "test_invalid_points_no_id_column.geojson"
ORIGINS_INVALID_DUPLICATE_IDS = (
    DATA_DIRECTORY / "test_invalid_points_duplicate_ids.geojson"
)
ORIGINS_VALID_IDS = DATA_DIRECTORY / "test_valid_points_data.geojson"
SINGLE_VALID_ORIGIN = DATA_DIRECTORY / "test_valid_single_point_data.geojson"
MULTIPLE_ORIGINS = DATA_DIRECTORY / "test_multiple_origins.geojson"


@pytest.fixture()
def origins_invalid_no_id():
    """Return a set of origins that has missing ID values."""
    origins = geopandas.read_file(ORIGINS_INVALID_NO_ID)
    yield origins


@pytest.fixture()
def origins_invalid_duplicate_ids():
    """Return a set of origins that has duplicate ID values."""
    # Since geopandas 1.0, it uses pyogrio in the background. pyogrio seems to
    # filter less of the underlying OGR warning messages than what fiona did.
    # Because of that, a warning message bubbles up that states non-unique IDs
    # were corrected (when they, factually, were not corrected)

    # for this fixture, we want to have non-unique values in the "id" column, so
    # let's ignore that warning

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=RuntimeWarning,
            message=(
                "Several features with id = 1 have been found. Altering it to be "
                "unique. This warning will not be emitted anymore for this layer"
            ),
        )
        origins = geopandas.read_file(ORIGINS_INVALID_DUPLICATE_IDS)
    yield origins


@pytest.fixture
def origin_point():
    """Return one origin point."""
    yield geopandas.read_file(SINGLE_VALID_ORIGIN)


@pytest.fixture()
def origins_valid_ids():
    """Return a set of origins that has valid ID values."""
    origins = geopandas.read_file(ORIGINS_VALID_IDS)
    yield origins


@pytest.fixture()
def multiple_origins():
    """Return a set of multiple origins."""
    origins = geopandas.read_file(MULTIPLE_ORIGINS)
    yield origins



#!/usr/bin/env python3

"""Python wrapper for the R5 routing analysis engine."""

__version__ = "1.0.2"


from .r5 import (
    DetailedItineraries,
    DetailedItinerariesComputer,
    Isochrones,
    RegionalTask,
    TransportMode,
    TransportNetwork,
    TravelTimeMatrix,
    TravelTimeMatrixComputer,
)

__all__ = [
    "DetailedItineraries",
    "DetailedItinerariesComputer",
    "Isochrones",
    "RegionalTask",
    "TransportMode",
    "TransportNetwork",
    "TravelTimeMatrix",
    "TravelTimeMatrixComputer",
    "__version__",
]



#!/usr/bin/env python3


"""Represent one leg of a direct mode (walk, cycle, car) trip."""


import datetime

import shapely

from .trip_leg import TripLeg


__all__ = ["DirectLeg"]


class DirectLeg(TripLeg):
    """Represent one leg of a public transport trip."""

    def __init__(self, transport_mode, street_segment):
        """
        Represent one leg of a public transport trip.

        Arguments
        =========
        transport_mode : r5py.TransportMode
            mode of transport this trip leg was travelled
        street_segment : com.conveyal.r5.profile.StreetSegment
            the leg’s data as output by R5’s `StreetRouter`
        """
        distance = street_segment.distance / 1000.0  # millimetres!
        travel_time = datetime.timedelta(seconds=street_segment.duration)
        geometry = shapely.from_wkt(str(street_segment.geometry))

        super().__init__(
            transport_mode=transport_mode,
            distance=distance,
            travel_time=travel_time,
            geometry=geometry,
        )



#!/usr/bin/env python3


"""A remote data set that is downloaded on demand."""


import hashlib
import pathlib
import warnings

from .config import Config
from .validating_requests_session import ValidatingRequestsSession


config = Config()


class SampleDataSet(pathlib.Path):
    """Data set that is downloaded and cached as needed."""

    # decide which kind of pathlib.Path we are (Windows, Unix, ...)
    # cf. https://stackoverflow.com/a/66613346/463864
    try:
        _flavour = type(pathlib.Path())._flavour
    except AttributeError:  # Python>=3.13
        pass

    _CACHE_DIR = pathlib.Path(config.CACHE_DIR) / "sampledata"

    def __new__(cls, remote_url, sha256_checksum):
        """Define a data set that is downloaded and cached on demand."""
        # pathlib.Path does everything in __new__, rather than __init__
        cached_path = cls._CACHE_DIR / pathlib.Path(remote_url).name
        return super().__new__(cls, cached_path)

    def __init__(self, remote_url, sha256_checksum, *args, **kwargs):
        """
        Define a data set that is downloaded and cached on demand.

        Arguments
        ---------
        remote_url : str
            source URL for this data set
        sha256_checksum : str
            checksum for this data set, using an SHA256 algorithm
        """
        cached_path = self._CACHE_DIR / pathlib.Path(remote_url).name

        try:  # Python>=3.12
            super().__init__(cached_path)
        except TypeError:
            super().__init__()

        self.remote_url = remote_url
        self.checksum = sha256_checksum
        self.cached_path = cached_path
        self._download_remote_file()

    def _download_remote_file(self):
        try:
            assert (
                hashlib.sha256(self.cached_path.read_bytes()).hexdigest()
                == self.checksum
            )
        except (AssertionError, FileNotFoundError):
            if config.arguments.verbose:
                warnings.warn(
                    f"First access to {pathlib.Path(self.remote_url).name}, "
                    "downloading remote file to local cache",
                    RuntimeWarning,
                )
            self.cached_path.parent.mkdir(exist_ok=True)
            with (
                ValidatingRequestsSession() as session,
                session.get(self.remote_url, self.checksum) as response,
            ):
                self.cached_path.write_bytes(response.content)



#!/usr/bin/env python3


"""Determine if an upstream R5 jar is used."""


import pytest


@pytest.fixture(scope="session")
def can_compute_detailed_route_geometries():
    """Determine if this R5 jar can output detailed route geometries"."""
    from r5py.util import start_jvm

    start_jvm()

    import com.conveyal.r5

    yield bool(com.conveyal.r5.transit.TransitLayer.SAVE_SHAPES)



#!/usr/bin/env python3

"""Calculate travel times between many origins and destinations."""

import copy

try:
    from warnings import deprecated
except ImportError:  # Python<=3.12
    from typing_extensions import deprecated

import pandas

from .base_travel_time_matrix import BaseTravelTimeMatrix
from ..util import start_jvm

import com.conveyal.r5


__all__ = ["TravelTimeMatrix", "TravelTimeMatrixComputer"]


start_jvm()


class TravelTimeMatrix(BaseTravelTimeMatrix):
    """Compute travel times between many origins and destinations."""

    def __init__(
        self,
        transport_network,
        origins=None,
        destinations=None,
        snap_to_network=False,
        **kwargs,
    ):
        """
        Compute travel times between many origins and destinations.

        ``r5py.TravelTimeMatrix`` are child classes of ``pandas.DataFrame`` and
        support all of their methods and properties,
        see https://pandas.pydata.org/docs/

        Arguments
        ---------
        transport_network : r5py.TransportNetwork | tuple(str, list(str), dict)
            The transport network to route on. This can either be a readily
            initialised r5py.TransportNetwork or a tuple of the parameters
            passed to ``TransportNetwork.__init__()``: the path to an OpenStreetMap
            extract in PBF format, a list of zero of more paths to GTFS transport
            schedule files, and a dict with ``build_config`` options.
        origins : geopandas.GeoDataFrame
            Places to find a route _from_
            Has to have a point geometry, and at least an `id` column
        destinations : geopandas.GeoDataFrame (optional)
            Places to find a route _to_
            Has to have a point geometry, and at least an `id` column
            If omitted, use same data set as for origins
        snap_to_network : bool or int, default False
            Should origin an destination points be snapped to the street network
            before routing? If `True`, the default search radius (defined in
            `com.conveyal.r5.streets.StreetLayer.LINK_RADIUS_METERS`) is used,
            if `int`, use `snap_to_network` meters as the search radius.
        **kwargs : mixed
            Any arguments than can be passed to r5py.RegionalTask:
            ``departure``, ``departure_time_window``, ``percentiles``, ``transport_modes``,
            ``access_modes``, ``egress_modes``, ``max_time``, ``max_time_walking``,
            ``max_time_cycling``, ``max_time_driving``, ``speed_cycling``, ``speed_walking``,
            ``max_public_transport_rides``, ``max_bicycle_traffic_stress``
        """
        super().__init__(
            transport_network,
            origins,
            destinations,
            snap_to_network,
            **kwargs,
        )
        data = self._compute()
        for column in data.columns:
            self[column] = data[column]
        del self.transport_network

    def _compute(self):
        """
        Compute travel times from all origins to all destinations.

        Returns
        -------
        pandas.DataFrame
            A data frame containing the columns ``from_id``, ``to_id``, and
            ``travel_time``, where ``travel_time`` is the median calculated
            travel time between ``from_id`` and ``to_id`` or ``numpy.nan``
            if no connection with the given parameters was found.
            If non-default ``percentiles`` were requested: one or more columns
            ``travel_time_p{:02d}`` representing the particular percentile of
            travel time.
        """
        self._prepare_origins_destinations()
        self.request.destinations = self.destinations

        od_matrix = pandas.concat(
            [self._travel_times_per_origin(from_id) for from_id in self.origins.id],
            ignore_index=True,
        )

        try:
            od_matrix = od_matrix.to_crs(self._origins_crs)
        except AttributeError:  # (not a GeoDataFrame)
            pass
        return od_matrix

    def _parse_results(self, from_id, results):
        """
        Parse the results of an R5 TravelTimeMatrix.

        Parse data as returned from `com.conveyal.r5.analyst.TravelTimeComputer.computeTravelTimes()`,
        cast data to Python types, and return as a `pandas.Dataframe`. Because of the way r5py
        and R5 interact, this parses the results of routing from one origin to many (all) destinations.

        Arguments
        ---------
        from_id : mixed
            The value of the ID column of the origin record to report on.
        results : `com.conveyal.r5.OneOriginResult` (Java object)

        Returns
        -------
        pandas.DataFrame
            A data frame containing the columns ``from_id``, ``to_id``, and
            ``travel_time``, where ``travel_time`` is the median calculated
            travel time between ``from_id`` and ``to_id`` or ``numpy.nan``
            if no connection with the given parameters was found.
            If non-default ``percentiles`` were requested: one or more columns
            ``travel_time_p{:02d}`` representing the particular percentile of
            travel time.
        """
        # First, create an empty DataFrame (this forces column types)
        od_matrix = pandas.DataFrame(
            {
                "from_id": pandas.Series(dtype=str),
                "to_id": pandas.Series(dtype=str),
            }
            | {
                f"travel_time_p{percentile:d}": pandas.Series(dtype=float)
                for percentile in self.request.percentiles
            }
        )

        # first assign columns with correct length (`to_id`),
        # only then fill `from_id` (it’s a scalar)
        od_matrix["to_id"] = self.destinations.id
        od_matrix["from_id"] = from_id

        for p, percentile in enumerate(self.request.percentiles):
            travel_times = results.travelTimes.getValues()[p]
            od_matrix[f"travel_time_p{percentile:d}"] = travel_times

        # rename percentile column if only median requested (the default)
        if self.request.percentiles == [50]:
            od_matrix = od_matrix.rename(columns={"travel_time_p50": "travel_time"})

        # R5’s NULL value is MAX_INT32
        od_matrix = self._fill_nulls(od_matrix)

        # re-index (and don’t keep the old index as a new column)
        od_matrix = od_matrix.reset_index(drop=True)

        return od_matrix

    def _travel_times_per_origin(self, from_id):
        request = copy.copy(self.request)
        request.origin = self.origins[self.origins.id == from_id].geometry.item()

        travel_time_computer = com.conveyal.r5.analyst.TravelTimeComputer(
            request, self.transport_network
        )
        results = travel_time_computer.computeTravelTimes()

        od_matrix = self._parse_results(from_id, results)

        return od_matrix


@deprecated(
    "Use `TravelTimeMatrix` instead, `TravelTimeMatrixComputer will be deprecated in a future release."
)
class TravelTimeMatrixComputer:
    """Compute travel times between many origins and destinations."""

    def __init__(self, *args, **kwargs):
        """Compute travel times between many origins and destinations."""
        self._ttm = TravelTimeMatrix(*args, **kwargs)

    def compute_travel_times(self):
        """
        Compute travel times from all origins to all destinations.

        Returns
        -------
        pandas.DataFrame
            A data frame containing the columns ``from_id``, ``to_id``, and
            ``travel_time``, where ``travel_time`` is the median calculated
            travel time between ``from_id`` and ``to_id`` or ``numpy.nan``
            if no connection with the given parameters was found.
            If non-default ``percentiles`` were requested: one or more columns
            ``travel_time_p{:02d}`` representing the particular percentile of
            travel time.
        """
        return self._ttm



#!/usr/bin/env python3

import pathlib
import shutil

import geopandas
import pytest
import pytest_lazy_fixtures
import shapely

from .temporary_directory import TemporaryDirectory

import r5py
import com.conveyal.r5
import java.time


class Test_TransportNetwork:
    def test_init_from_files_and_dir_cover_same_extent(
        self,
        transport_network_from_test_files,
        transport_network_from_test_directory,
    ):
        assert (
            transport_network_from_test_files._transport_network.getEnvelope()
            == transport_network_from_test_directory._transport_network.getEnvelope()
        )

    @pytest.mark.parametrize(
        ["transport_network"],
        [
            (pytest_lazy_fixtures.lf("transport_network_from_test_files"),),
            (pytest_lazy_fixtures.lf("transport_network_from_test_directory"),),
        ],
    )
    def test_transport_network_java_object(self, transport_network):
        assert isinstance(
            transport_network._transport_network,
            com.conveyal.r5.transit.TransportNetwork,
        )

    @pytest.mark.parametrize(
        ["transport_network"],
        [
            (pytest_lazy_fixtures.lf("transport_network_from_test_files"),),
            (pytest_lazy_fixtures.lf("transport_network_from_test_directory"),),
        ],
    )
    def test_context(self, transport_network):
        with transport_network as tn:
            assert isinstance(tn, r5py.TransportNetwork)
            assert tn == transport_network

    @pytest.mark.parametrize(
        ["transport_network"],
        [
            (pytest_lazy_fixtures.lf("transport_network_from_test_files"),),
            (pytest_lazy_fixtures.lf("transport_network_from_test_directory"),),
        ],
    )
    def test_linkage_cache_java_object(self, transport_network):
        assert isinstance(
            transport_network.linkage_cache, com.conveyal.r5.analyst.LinkageCache
        )

    @pytest.mark.parametrize(
        ["transport_network"],
        [
            (pytest_lazy_fixtures.lf("transport_network_from_test_files"),),
            (pytest_lazy_fixtures.lf("transport_network_from_test_directory"),),
        ],
    )
    def test_street_layer(self, transport_network):
        assert isinstance(transport_network.street_layer, r5py.r5.StreetLayer)

    @pytest.mark.parametrize(
        ["transport_network"],
        [
            (pytest_lazy_fixtures.lf("transport_network_from_test_files"),),
            (pytest_lazy_fixtures.lf("transport_network_from_test_directory"),),
        ],
    )
    def test_timezone(self, transport_network, gtfs_timezone_helsinki):
        assert isinstance(transport_network.timezone, java.time.ZoneId)
        assert transport_network.timezone.toString() == gtfs_timezone_helsinki

    def test_fromdirectory_multiple_osm_files(self, transport_network_files_tuple):
        # try to create transport network from a directory, in which
        # more than one osm file is found
        osm, gtfs = transport_network_files_tuple
        with TemporaryDirectory() as temp_directory:
            temp_directory = pathlib.Path(temp_directory)
            shutil.copy(osm, temp_directory / "first.osm.pbf")
            shutil.copy(osm, temp_directory / "second.osm.pbf")
            with pytest.warns(RuntimeWarning):
                transport_network = r5py.TransportNetwork.from_directory(  # noqa: F841
                    temp_directory
                )

    def test_fromdirectory_no_osm_files(self):
        # try to create transport network from a directory without osm file
        with TemporaryDirectory() as temp_directory:
            temp_directory = pathlib.Path(temp_directory)
            with pytest.raises(FileNotFoundError):
                transport_network = r5py.TransportNetwork.from_directory(  # noqa: F841
                    temp_directory
                )

    def test_snap_to_network(
        self,
        transport_network,
        population_grid_points,
        snapped_population_grid_points,
    ):
        snapped = transport_network.snap_to_network(population_grid_points.geometry)
        geopandas.testing.assert_geoseries_equal(
            snapped.geometry, snapped_population_grid_points.geometry
        )

    def test_snap_to_network_with_unsnappable_points(
        self,
        transport_network,
        unsnappable_points,
    ):
        snapped = transport_network.snap_to_network(unsnappable_points.geometry)
        assert snapped.geometry.unique() == [shapely.Point()]

    def test_failed_symlink(self, transport_network_files_tuple, monkeypatch):
        def _symlink_to(*args, **kwargs):
            raise OSError

        monkeypatch.setattr(pathlib.Path, "symlink_to", _symlink_to)

        transport_network = r5py.TransportNetwork(*transport_network_files_tuple)
        del transport_network

    @pytest.mark.parametrize(
        ["osm_pbf_type", "gtfs_type", "gtfs_is_list"],
        [
            (pathlib.Path, pathlib.Path, False),
            (pathlib.Path, pathlib.Path, True),
            (pathlib.Path, str, False),
            (pathlib.Path, str, True),
            (str, pathlib.Path, False),
            (str, pathlib.Path, True),
            (str, str, False),
            (str, str, True),
        ],
    )
    def test_argument_types(
        self,
        osm_pbf_file_path,
        gtfs_file_path,
        osm_pbf_type,
        gtfs_type,
        gtfs_is_list,
    ):
        osm_pbf = osm_pbf_type(osm_pbf_file_path)
        gtfs = gtfs_type(gtfs_file_path)
        if gtfs_is_list:
            gtfs = [gtfs]
        transport_network = r5py.TransportNetwork(osm_pbf, gtfs)
        del transport_network



#!/usr/bin/env python3


"""Fixtures related to regional tasks and routing parameters."""


import datetime

import pytest


@pytest.fixture(scope="session")
def departure_datetime():
    """Return the departure time to run tests for."""
    yield datetime.datetime(2022, 2, 22, 8, 30)


@pytest.fixture
def regional_task(
    population_grid_points,
    transport_network,
    departure_datetime,
):
    """Return an initialised `r5py.RegionalTask`."""
    import r5py

    regional_task = r5py.RegionalTask(
        transport_network,
        population_grid_points.at[1, "geometry"],
        population_grid_points,
        departure=departure_datetime,
    )
    yield regional_task



#!/usr/bin/env python3


import pathlib
import sys

import pytest

from r5py.util.config import Config
from r5py.util.sample_data_set import SampleDataSet


class TestSampleDataSet:
    def test_data_set_with_verbose(self, sample_data_set_url, sample_data_set_sha256):
        sys.argv.extend(
            [
                "--verbose",
            ]
        )

        # make sure the cached file does not exist
        try:
            (
                pathlib.Path(Config().CACHE_DIR)
                / "sampledata"
                / pathlib.Path(sample_data_set_url).name
            ).unlink()
        except FileNotFoundError:
            pass

        with pytest.warns(
            RuntimeWarning,
            match="First access to .*, downloading remote file to local cache",
        ):
            data_set = SampleDataSet(sample_data_set_url, sample_data_set_sha256)
            assert data_set.exists()

        data_set.unlink()
        sys.argv = sys.argv[:-1]

    def test_data_set(self, sample_data_set_url, sample_data_set_sha256):
        data_set = SampleDataSet(sample_data_set_url, sample_data_set_sha256)
        assert data_set.exists()
        data_set.unlink()

    def test_data_set_invalid_hash(self, sample_data_set_url, sample_data_set_sha256):
        data_set = SampleDataSet(sample_data_set_url, sample_data_set_sha256)
        data_set.write_text("foobar")  # change file, invalidate checksum
        del data_set

        # try again:
        data_set = SampleDataSet(sample_data_set_url, sample_data_set_sha256)
        assert data_set.exists()
        data_set.unlink()



#!/usr/bin/env python3

import datetime

import pytest
import shapely

import r5py


class TestTripLeg:
    def test_repr(self):
        trip_leg = r5py.r5.trip_leg.TripLeg(
            transport_mode=r5py.TransportMode.TRANSIT,
            departure_time=datetime.datetime(2023, 4, 25, 15, 30),
            distance=12.67,
            travel_time=datetime.timedelta(minutes=10),
            wait_time=datetime.timedelta(minutes=2),
            geometry=shapely.LineString([[0, 0], [1, 1], [2, 2]]),
        )
        assert (
            repr(trip_leg)
            == "<TripLeg: TransportMode.TRANSIT, 12.67m, 600.0s, (0.0, 0.0) -> (2.0, 2.0)>"
        )

        trip_leg = r5py.r5.trip_leg.TripLeg()
        assert repr(trip_leg) == "<TripLeg>"

    def test_add(self):
        trip_leg1 = r5py.r5.trip_leg.TripLeg(
            r5py.TransportMode.TRANSIT,
            datetime.datetime(2023, 4, 25, 15, 30),
            12.34,
            datetime.timedelta(minutes=10),
            datetime.timedelta(minutes=2),
            "56A",
            shapely.LineString([[0, 0], [1, 1], [2, 2]]),
        )
        trip_leg2 = r5py.r5.trip_leg.TripLeg(
            r5py.TransportMode.TRANSIT,
            datetime.datetime(2023, 4, 25, 15, 30),
            98.76,
            datetime.timedelta(minutes=30),
            datetime.timedelta(minutes=5),
            "54A",
            shapely.LineString([[2, 2], [34, 34]]),
        )

        trip = trip_leg1 + trip_leg2
        assert isinstance(trip, r5py.r5.trip.Trip)
        assert trip.legs == [trip_leg1, trip_leg2]

        trip = trip_leg2 + trip_leg1
        assert isinstance(trip, r5py.r5.trip.Trip)
        assert trip.legs == [trip_leg2, trip_leg1]

        trip = sum([trip_leg1, trip_leg2])
        assert isinstance(trip, r5py.r5.trip.Trip)
        assert trip.legs == [trip_leg1, trip_leg2]

        trip = r5py.r5.trip.Trip([trip_leg1])
        trip += trip_leg2
        assert isinstance(trip, r5py.r5.trip.Trip)
        assert trip.legs == [trip_leg1, trip_leg2]

        trip = r5py.r5.trip.Trip([trip_leg1])
        trip = trip_leg2 + trip
        assert isinstance(trip, r5py.r5.trip.Trip)
        assert trip.legs == [trip_leg2, trip_leg1]

    @pytest.mark.parametrize(
        ["invalid_other"],
        [
            (123.0,),
            (1,),
            ("asdfasdf",),
            ({},),
        ],
    )
    def test_add_invalid_type(self, invalid_other):
        trip_leg = r5py.r5.trip_leg.TripLeg(
            r5py.TransportMode.TRANSIT,
            datetime.datetime(2023, 4, 25, 15, 30),
            98.76,
            datetime.timedelta(minutes=30),
            datetime.timedelta(minutes=5),
            "54A",
            shapely.LineString([[2, 2], [34, 34]]),
        )
        with pytest.raises(TypeError, match="unsupported operand type"):
            _ = trip_leg + invalid_other

        with pytest.raises(TypeError, match="unsupported operand type"):
            _ = invalid_other + trip_leg

    def test_as_table_row(self):
        trip_leg = r5py.r5.trip_leg.TripLeg(
            transport_mode=r5py.TransportMode.TRANSIT,
            departure_time=datetime.datetime(2023, 4, 25, 15, 30),
            distance=12.67,
            travel_time=datetime.timedelta(minutes=10),
            wait_time=datetime.timedelta(minutes=2),
            feed="GTFS",
            agency_id="HSR",
            route_id="48A",
            start_stop_id="20302",
            end_stop_id="18803",
            geometry=shapely.LineString([[0, 0], [1, 1], [2, 2]]),
        )
        assert trip_leg.as_table_row() == [
            r5py.TransportMode.TRANSIT,
            datetime.datetime(2023, 4, 25, 15, 30),
            12.67,
            datetime.timedelta(minutes=10),
            datetime.timedelta(minutes=2),
            "GTFS",
            "HSR",
            "48A",
            "20302",
            "18803",
            shapely.LineString([[0, 0], [1, 1], [2, 2]]),
        ]

    def test_compare(self):
        trip_leg1 = r5py.r5.trip_leg.TripLeg(
            r5py.TransportMode.TRANSIT,
            datetime.datetime(2023, 4, 25, 15, 30),
            12.34,
            datetime.timedelta(minutes=10),
            datetime.timedelta(minutes=2),
            "56A",
            shapely.LineString([[0, 0], [1, 1], [2, 2]]),
        )
        trip_leg2 = r5py.r5.trip_leg.TripLeg(
            r5py.TransportMode.TRANSIT,
            datetime.datetime(2023, 4, 25, 15, 30),
            98.76,
            datetime.timedelta(minutes=30),
            datetime.timedelta(minutes=5),
            "54A",
            shapely.LineString([[2, 2], [34, 34]]),
        )

        assert trip_leg1 < trip_leg2
        assert trip_leg1 <= trip_leg2
        assert trip_leg2 > trip_leg1
        assert trip_leg2 >= trip_leg1



#!/usr/bin/env python3

"""Calculate travel times between many origins and destinations."""

import math
import multiprocessing
import warnings

import geopandas
import numpy
import shapely

from ..util import check_od_data_set, Config
from .regional_task import RegionalTask
from .transport_network import TransportNetwork


__all__ = ["BaseTravelTimeMatrix"]


# R5 fills cut-off (NULL) values with MAX_INT32
MAX_INT32 = (2**31) - 1

# how many (Python) threads to start
# (they still run many Java threads, so be careful what you wish for ;) )
# TODO: benchmark the optimal number of threads
NUM_THREADS = math.ceil(multiprocessing.cpu_count() * 0.5)


class BaseTravelTimeMatrix(geopandas.GeoDataFrame):
    """Base class for travel time computers between many origins and destinations."""

    MAX_INT32 = MAX_INT32

    NUM_THREADS = NUM_THREADS

    _r5py_attributes = [
        "_destinations",
        "_destinations_crs",
        "_origins",
        "_origins_crs",
        "destinations",
        "origins",
        "request",
        "snap_to_network",
        "transport_network",
        "verbose",
    ]

    def __init__(
        self,
        transport_network,
        origins=None,
        destinations=None,
        snap_to_network=False,
        **kwargs,
    ):
        """
        Compute travel times between many origins and destinations.

        Arguments
        ---------
        transport_network : r5py.TransportNetwork | tuple(str, list(str), dict)
            The transport network to route on. This can either be a readily
            initialised r5py.TransportNetwork or a tuple of the parameters
            passed to ``TransportNetwork.__init__()``: the path to an OpenStreetMap
            extract in PBF format, a list of zero of more paths to GTFS transport
            schedule files, and a dict with ``build_config`` options.
        origins : geopandas.GeoDataFrame
            Places to find a route _from_
            Has to have a point geometry, and at least an `id` column
        destinations : geopandas.GeoDataFrame (optional)
            Places to find a route _to_
            Has to have a point geometry, and at least an `id` column
            If omitted, use same data set as for origins
        snap_to_network : bool or int, default False
            Should origin an destination points be snapped to the street network
            before routing? If `True`, the default search radius (defined in
            `com.conveyal.r5.streets.StreetLayer.LINK_RADIUS_METERS`) is used,
            if `int`, use `snap_to_network` meters as the search radius.
        **kwargs : mixed
            Any arguments than can be passed to r5py.RegionalTask:
            ``departure``, ``departure_time_window``, ``percentiles``, ``transport_modes``,
            ``access_modes``, ``egress_modes``, ``max_time``, ``max_time_walking``,
            ``max_time_cycling``, ``max_time_driving``, ``speed_cycling``, ``speed_walking``,
            ``max_public_transport_rides``, ``max_bicycle_traffic_stress``
        """
        geopandas.GeoDataFrame.__init__(self)

        if not isinstance(transport_network, TransportNetwork):
            transport_network = TransportNetwork(*transport_network)
        self.transport_network = transport_network

        self.snap_to_network = snap_to_network

        self.origins = origins
        self.destinations = destinations

        self.request = RegionalTask(
            transport_network,
            origin=None,
            destinations=None,
            **kwargs,
        )

        self.verbose = Config().arguments.verbose

    def __setattr__(self, attr, val):
        """Catch our own attributes here so we don’t mess with (geo)pandas columns."""
        if attr in self._r5py_attributes:
            object.__setattr__(self, attr, val)
        else:
            super().__setattr__(attr, val)

    @property
    def destinations(self):
        """The destinations of this travel time matrix (`geopandas.GeoDataFrame`)."""
        return self._destinations

    @destinations.setter
    def destinations(self, destinations):
        if destinations is not None:
            check_od_data_set(destinations)
            self._destinations_crs = destinations.crs
            self._destinations = destinations.to_crs("EPSG:4326").copy()

    def _fill_nulls(self, data_set):
        """
        Fill NULL values in a data set returned from R5.

        R5 uses `MAX_INT32` as a marker for NULL values, this function
        replaces those values in `data_set` with `numpy.nan`

        Arguments
        ---------
        data_set : pandas.DataFrame
            Data frame in which NULLs are represented by MAX_INT32

        Returns
        -------
        pandas.DataFrame
            Data frame in which all MAX_INT32 have been replaced by `numpy.nan`.
        """
        return data_set.map(lambda x: numpy.nan if x == MAX_INT32 else x)

    def _prepare_origins_destinations(self):
        """Make sure we received enough information to route from origins to destinations."""
        try:
            self.origins
        except AttributeError as exception:
            raise ValueError("No routing origins defined") from exception

        try:
            self.destinations
            assert self.destinations is not None
        except (AssertionError, AttributeError):
            self.destinations = self.origins.copy()
            if self.verbose:
                warnings.warn(
                    "No routing destinations defined, using origins as destinations, too.",
                    RuntimeWarning,
                )

        if self.snap_to_network:
            for which_end in ("origins", "destinations"):
                points = getattr(self, f"_{which_end}")
                points.geometry = self.transport_network.snap_to_network(
                    points.geometry
                )
                if len(points[points.geometry == shapely.Point()]):
                    # if there are origins/destinations for which no snapped point could be found
                    points = points[points.geometry != shapely.Point()]
                    warnings.warn(
                        f"Some {which_end[:-1]} points could not be snapped to the street network",
                        RuntimeWarning,
                    )

                    if points.empty:
                        raise ValueError(
                            f"After snapping, no valid {which_end[:-1]} points remain"
                        )

                setattr(self, f"_{which_end}", points.copy())

    @property
    def origins(self):
        """The origins of this travel time matrix (`geopandas.GeoDataFrame`)."""
        return self._origins

    @origins.setter
    def origins(self, origins):
        if origins is not None:
            check_od_data_set(origins)
            self._origins_crs = origins.crs
            self._origins = origins.to_crs("EPSG:4326").copy()



#!/usr/bin/env python3


import pytest

from r5py.r5.transit_layer import TransitLayer


class TestTransitLayer:
    def test_uninitialised_transit_layer_start_date(self):
        transit_layer = TransitLayer()
        with pytest.raises((AttributeError, ValueError)):
            assert transit_layer.start_date < transit_layer.end_date

    def test_uninitialised_transit_layer_end_date(self):
        transit_layer = TransitLayer()
        with pytest.raises((AttributeError, ValueError)):
            assert transit_layer.end_date > transit_layer.start_date



#!/usr/bin/env python3

"""Create a hash sum of a file."""


import hashlib
import pathlib


__all__ = ["FileDigest"]


BUFFER_SIZE = 64 * 1024


class FileDigest(str):
    """Create a hash sum of a file."""

    def __new__(cls, input_file, digest="blake2s"):
        """
        Create a hash sum of a file.

        Arguments
        ---------
        input_file : pathlib.Path | str
            for which file to compute a hash digest
        digest : str | func
            name of hash algorithm (s.
            https://docs.python.org/3/library/hashlib.html) or function that
            returns a hash sum
        """
        input_file = pathlib.Path(input_file)
        try:
            with input_file.open("rb") as f:
                hashdigest = hashlib.file_digest(f, digest)
        except AttributeError:  # Python<=3.10
            hashdigest = hashlib.new(digest)
            with input_file.open("rb") as f:
                while data := f.read(BUFFER_SIZE):
                    hashdigest.update(data)

        return hashdigest.hexdigest()



#!/usr/bin/env python3

import pytest

import r5py
import com.conveyal.r5


class TestJavaCasting:
    @pytest.mark.parametrize(
        ["python_object", "java_class"],
        [
            (r5py.TransportMode.AIR, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.BUS, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.CABLE_CAR, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.FERRY, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.FUNICULAR, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.GONDOLA, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.RAIL, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.SUBWAY, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.TRAM, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.TRANSIT, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.BICYCLE, com.conveyal.r5.profile.StreetMode),
            (r5py.TransportMode.CAR, com.conveyal.r5.profile.StreetMode),
            (r5py.TransportMode.WALK, com.conveyal.r5.profile.StreetMode),
            (r5py.TransportMode.BICYCLE, com.conveyal.r5.api.util.LegMode),
            (r5py.TransportMode.CAR, com.conveyal.r5.api.util.LegMode),
            (r5py.TransportMode.WALK, com.conveyal.r5.api.util.LegMode),
            (r5py.TransportMode.BICYCLE_RENT, com.conveyal.r5.api.util.LegMode),
            (r5py.TransportMode.CAR_PARK, com.conveyal.r5.api.util.LegMode),
        ],
    )
    def test_transport_modes_cast(self, python_object, java_class):
        # see https://jpype.readthedocs.io/en/latest/quickguide.html#classes-objects
        _ = (java_class) @ python_object

    @pytest.mark.parametrize(
        ["python_object", "java_class"],
        [
            (r5py.TransportMode.WALK, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.CAR, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.BICYCLE_RENT, com.conveyal.r5.api.util.TransitModes),
            (r5py.TransportMode.TRAM, com.conveyal.r5.profile.StreetMode),
            (r5py.TransportMode.BUS, com.conveyal.r5.profile.StreetMode),
            (r5py.TransportMode.CAR_PARK, com.conveyal.r5.profile.StreetMode),
            (r5py.TransportMode.TRANSIT, com.conveyal.r5.api.util.LegMode),
            (r5py.TransportMode.FUNICULAR, com.conveyal.r5.api.util.LegMode),
            (r5py.TransportMode.GONDOLA, com.conveyal.r5.api.util.LegMode),
        ],
    )
    def test_transport_modes_invalid_cast(self, python_object, java_class):
        # see https://jpype.readthedocs.io/en/latest/quickguide.html#classes-objects
        with pytest.raises(ValueError, match="is not a valid R5"):
            _ = (java_class) @ python_object



#!/usr/bin/env python3


import pytest

import hashlib
import os
import pathlib
import sys

from r5py.util.classpath import find_r5_classpath
from r5py.util.config import Config
from r5py.util.exceptions import UnexpectedClasspathSchema


class TestClassPath:
    @pytest.mark.skipif(
        sys.platform == "win32", reason="Windows cannot delete jar while JVM runs"
    )
    def test_download_classpath_with_verbose(
        self, r5_jar_url, r5_jar_sha256, r5_jar_cached, r5_jar_cached_invalid
    ):
        sys.argv.extend(
            [
                "--verbose",
                "--r5-classpath",
                r5_jar_cached_invalid,
            ]
        )

        try:
            pathlib.Path(r5_jar_cached).unlink()  # delete cached jar!
        except FileNotFoundError:
            pass

        with pytest.warns(RuntimeWarning, match="Could not find R5 jar"):
            r5_classpath = find_r5_classpath(Config().arguments)
        with open(r5_classpath, "rb") as r5_jar:
            digest = hashlib.sha256(r5_jar.read()).hexdigest()
        assert digest == r5_jar_sha256

        sys.argv = sys.argv[:-3]

    def test_use_classpath_from_local_file(
        self, r5_jar_url, r5_jar_sha256, r5_jar_cached
    ):
        # run `find_r5_classpath` once in order to download the jar into cache
        find_r5_classpath(Config().arguments)

        sys.argv.extend(["--r5-classpath", r5_jar_cached])

        r5_classpath = find_r5_classpath(Config().arguments)

        with open(r5_classpath, "rb") as r5_jar:
            digest = hashlib.sha256(r5_jar.read()).hexdigest()

        assert digest == r5_jar_sha256
        assert r5_classpath == r5_jar_cached

        sys.argv = sys.argv[:-2]

    def test_use_classpath_from_local_uri(
        self, r5_jar_url, r5_jar_sha256, r5_jar_cached
    ):
        # run `find_r5_classpath` once in order to download the jar into cache
        find_r5_classpath(Config().arguments)

        sys.argv.extend(["--r5-classpath", f"file://{r5_jar_cached}"])

        r5_classpath = find_r5_classpath(Config().arguments)

        with open(r5_classpath, "rb") as r5_jar:
            digest = hashlib.sha256(r5_jar.read()).hexdigest()

        assert digest == r5_jar_sha256
        assert r5_classpath == r5_jar_cached

        sys.argv = sys.argv[:-2]

    def test_use_classpath_from_remote_uri(
        self, r5_jar_url, r5_jar_sha256, r5_jar_cached
    ):
        sys.argv.extend(["--r5-classpath", r5_jar_url])

        r5_classpath = find_r5_classpath(Config().arguments)

        with open(r5_classpath, "rb") as r5_jar:
            digest = hashlib.sha256(r5_jar.read()).hexdigest()

        assert digest == r5_jar_sha256
        assert r5_classpath == r5_jar_cached

        sys.argv = sys.argv[:-2]

    def test_use_classpath_from_invalid_uri(self):
        sys.argv.extend(["--r5-classpath", "invalid://schema/and/path"])

        with pytest.raises(UnexpectedClasspathSchema):
            _ = find_r5_classpath(Config().arguments)

        sys.argv = sys.argv[:-2]

    def test_find_classpath_download(
        self, r5_jar_url, r5_jar_sha256, r5_jar_cached_invalid
    ):
        sys.argv.extend(["--r5-classpath", r5_jar_cached_invalid])
        r5_classpath = find_r5_classpath(Config().arguments)
        with open(r5_classpath, "rb") as r5_jar:
            digest = hashlib.sha256(r5_jar.read()).hexdigest()
        assert digest == r5_jar_sha256

        sys.argv = sys.argv[:-2]

    @pytest.mark.skipif(
        sys.platform == "win32", reason="No signal chaining library for Windows"
    )
    def test_signal_chaining(self):
        if sys.platform == "linux":
            assert "LD_PRELOAD" in os.environ
            assert pathlib.Path(os.environ["LD_PRELOAD"]).exists()
        elif sys.platform == "darwin":
            assert "DYLD_INSERT_LIBRARIES" in os.environ
            assert pathlib.Path(os.environ["DYLD_INSERT_LIBRARIES"]).exists()



#!/usr/bin/env python3


import pytest

from r5py.util.memory_footprint import (
    _get_max_memory,
    _interpret_power_of_two_units,
    _parse_value_and_unit,
    _share_of_ram,
)
import psutil


class TestMemoryFootprint:
    @pytest.mark.parametrize(
        ["memory_input", "expected"],
        [
            ("5G", (5, "G")),
            ("1000M", (1000, "M")),
            ("1T", (1, "T")),
            ("1073741824K", (1073741824, "K")),
            ("1073741824", (1073741824, None)),
        ],
    )
    def test_parse_value_and_unit(self, memory_input, expected):
        assert _parse_value_and_unit(memory_input) == expected

    def test_incorrect_memory_input(self):
        # Totally incorrect input
        value = "incorrect_value"
        with pytest.raises(ValueError, match="Could not interpret .*"):
            _get_max_memory(value)

        # Incorrect suffix
        value = "1000J"
        with pytest.raises(ValueError, match="Could not interpret unit .*"):
            _get_max_memory(value)

    @pytest.mark.parametrize(
        ["value", "unit", "expected"],
        [
            (5, "G", 5368709120),
            (1.5, "G", 1610612736),
            (1000, "M", 1048576000),
            (1, "T", 1099511627776),
            (1048576, "K", 1073741824),
            (1073741824, None, 1073741824),
        ],
    )
    def test_interpret_power_of_two_units(self, value, unit, expected):
        assert _interpret_power_of_two_units(value, unit) == expected

    def test_lower_than_allowed_memory_input(self):
        value = "1"
        with pytest.warns(
            RuntimeWarning, match="Requested maximum JVM heap size is too low for R5"
        ):
            assert _get_max_memory(value) == 209715200

    @pytest.mark.parametrize(
        ["share", "leave_at_least", "expected"],
        [
            (0.8, 0, psutil.virtual_memory().total * 0.8),
            (0.1, 0, psutil.virtual_memory().total * 0.1),
            (1.0, 2000, psutil.virtual_memory().total - 2000),
        ],
    )
    def test_share_of_ram_leaving_zero(self, share, leave_at_least, expected):
        # NOTE: this can never match exactly, as memory is consumed
        # in the computation of the share, and most likely
        # by other processes running at the same time.

        # Let’s try to see whether we get to within 100MiB
        assert _share_of_ram(share, leave_at_least) == pytest.approx(
            expected, 100 * 1024**2
        )

    # There remain two tricky cases untested:
    #    - share_of_ram() with leave_at_least > 0:
    #      very tricky to test without reproducing the to-be-tested
    #      code 1:1, dynamically depending on available RAM
    #    - share_of_ram() with a leave_at_least > (total - share).
    #      Even more complicated ;)



#!/usr/bin/env python3


"""Normalise some environment variables that might not always get set."""


import os
import pathlib


# if a readthedocs runner uses a conda environment, it fails to
# properly initialise the JAVA_HOME and PROJ_LIB environment variables
#
# this might happen on other installation, so let’s keep this as general
# as possible.
#
# As readthedocs also does not export CONDA_PREFIX, we first reconstruct
# it from CONDA_ENVS_PATH and CONDA_DEFAULT_ENV
if (
    "CONDA_PREFIX" not in os.environ
    and "CONDA_DEFAULT_ENV" in os.environ
    and "CONDA_ENVS_PATH" in os.environ
):
    os.environ["CONDA_PREFIX"] = str(
        pathlib.Path(os.environ["CONDA_ENVS_PATH"]) / os.environ["CONDA_DEFAULT_ENV"]
    )
if "JAVA_HOME" not in os.environ and "CONDA_PREFIX" in os.environ:
    os.environ["JAVA_HOME"] = str(
        pathlib.Path(os.environ["CONDA_PREFIX"]) / "lib" / "jvm"
    )
if "PROJ_LIB" not in os.environ and "CONDA_PREFIX" in os.environ:
    os.environ["PROJ_LIB"] = str(
        pathlib.Path(os.environ["CONDA_PREFIX"]) / "share" / "proj"
    )



#!/usr/bin/env python3


"""Find the most appropriate equidistant (UTM) reference system for an extent."""


import pyproj
import shapely

from .exceptions import UnexpectedCrsError


FALLBACK_CRS = 3857
DATUM_NAME = "WGS 84"
VERY_SMALL_BUFFER_SIZE = 0.001


class GoodEnoughEquidistantCrs(pyproj.CRS):
    """
    Find the most appropriate UTM reference system for the current extent.

    (We need this to be able to calculate lengths in meters.
    Results don’t have to be perfect, so also the neighbouring UTM grid will do.)
    """

    def __new__(cls, extent):
        """
        Find the most appropriate UTM reference system for the current extent.

        (We need this to be able to calculate lengths in meters.
        Results don’t have to be perfect, so also the neighbouring UTM grid will do.)

        Arguments
        ---------
        extent: shapely.Geometry
            The geographical extent for which to find an equidistant reference
            system, in `EPSG:4326`
        """
        if GoodEnoughEquidistantCrs._is_plausible_in_epsg4326(extent):
            # default CRS in case we do not find any better match
            crs = pyproj.CRS.from_epsg(FALLBACK_CRS)

            # buffer extent (so everything is a polygon)
            extent = extent.buffer(VERY_SMALL_BUFFER_SIZE)

            crsinfo = pyproj.database.query_utm_crs_info(
                datum_name=DATUM_NAME,
                area_of_interest=pyproj.aoi.AreaOfInterest(*extent.bounds),
            )
            for candidate_crs in crsinfo:
                area_of_use = shapely.box(*candidate_crs.area_of_use.bounds)
                coverage = shapely.intersection(extent, area_of_use).area / extent.area

                if coverage > 0.5:
                    # more than half of extent covered by crs’ area of use
                    # -> good enough
                    crs = pyproj.CRS.from_authority(
                        candidate_crs.auth_name, candidate_crs.code
                    )
                    break

            return crs

        else:
            raise UnexpectedCrsError("`extent` does not seem to be in `EPSG:4326`")

    @staticmethod
    def _is_plausible_in_epsg4326(geometry):
        try:
            minx, miny, maxx, maxy = geometry.bounds
            assert -180 <= minx <= maxx <= 180
            assert -90 <= miny <= maxy <= 90
            return True
        except AssertionError:
            return False



#!/usr/bin/env python3

"""Wraps a com.conveyal.r5.analyst.scenario.Scenario."""

import jpype

from ..util import start_jvm

import com.conveyal.r5


__all__ = ["Scenario"]


start_jvm()


class Scenario:
    """Wrap a com.conveyal.r5.analyst.scenario.Scenario."""

    def __init__(self):
        """Initialise a most simple Scenario."""
        scenario = com.conveyal.r5.analyst.scenario.Scenario()
        scenario.id = "id"
        self._scenario = scenario

    @property
    def id(self):
        """Retrieve the Java Scenario’s instance’s ID."""
        return self._scenario.id


@jpype._jcustomizer.JConversion(
    "com.conveyal.r5.analyst.scenario.Scenario", exact=Scenario
)
def _cast_Scenario(java_class, object_):
    return object_._scenario



#!/usr/bin/env python3

"""Make sure R5 is in the class path, download it if not."""

import hashlib
import pathlib
import string
import urllib.parse
import warnings

import requests

from .config import Config
from .exceptions import UnexpectedClasspathSchema
from .validating_requests_session import ValidatingRequestsSession
from .warnings import R5pyWarning


# update these to use a newer R5 version if no R5 available locally
R5_JAR_URL = (
    "https://github.com/r5py/r5/releases/download/v7.3-r5py/r5-v7.3-r5py-all.jar"
)
R5_JAR_SHA256 = "cb1ccad370757ba229cf17f1bedc9549ff5d77fdbb44b7a3058104fe1f243f53"
# ---


__all__ = ["R5_CLASSPATH"]


config = Config()

config.argparser.add(
    "-r",
    "--r5-classpath",
    help="R5’s class path, can point to r5-all.jar",
    default="",
)


def find_r5_classpath(arguments):
    r5_classpath = None

    if arguments.r5_classpath:
        schema, *_ = urllib.parse.urlparse(arguments.r5_classpath)

        # fmt: off
        if (
            schema in ("file", "")
            or (len(schema) == 1 and schema in string.ascii_letters)  # windows drive letter
        ):
            # fmt: on
            if pathlib.Path(arguments.r5_classpath).exists():
                r5_classpath = arguments.r5_classpath

        elif schema in ("https", "http"):
            r5_classpath = config.CACHE_DIR / pathlib.Path(arguments.r5_classpath).name
            with requests.get(arguments.r5_classpath) as response:
                r5_classpath.write_bytes(response.content)
            r5_classpath = str(r5_classpath)

        else:
            raise UnexpectedClasspathSchema(
                f"Could not parse `r5_classpath`: "
                f"schema {schema}:// is not supported"
            )

    if r5_classpath is None:
        r5_classpath = str(config.CACHE_DIR / pathlib.Path(R5_JAR_URL).name)
        try:
            with open(r5_classpath, "rb") as jar:
                assert hashlib.sha256(jar.read()).hexdigest() == R5_JAR_SHA256
        except (AssertionError, FileNotFoundError):
            if arguments.verbose:
                warnings.warn(
                    "Could not find R5 jar, trying to download it from upstream",
                    R5pyWarning,
                )
            with (
                ValidatingRequestsSession() as session,
                session.get(R5_JAR_URL, R5_JAR_SHA256) as response,
                open(r5_classpath, "wb") as jar,
            ):
                jar.write(response.content)
            if arguments.verbose:
                warnings.warn(
                    f"Successfully downloaded {pathlib.Path(R5_JAR_URL).name}",
                    R5pyWarning,
                )

    return r5_classpath


R5_CLASSPATH = find_r5_classpath(config.arguments)
